/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package edu.mayo.informatics.lexgrid.convert.emfConversions.hl7;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Hashtable;

import org.LexGrid.LexBIG.Preferences.loader.LoadPreferences.LoaderPreferences;
import org.LexGrid.emf.codingSchemes.CodingScheme;
import org.LexGrid.emf.commonTypes.CommontypesFactory;
import org.LexGrid.emf.commonTypes.Property;
import org.LexGrid.emf.commonTypes.PropertyQualifier;
import org.LexGrid.emf.commonTypes.Source;
import org.LexGrid.emf.commonTypes.Text;
import org.LexGrid.emf.concepts.Concept;
import org.LexGrid.emf.concepts.ConceptsFactory;
import org.LexGrid.emf.concepts.Entities;
import org.LexGrid.emf.concepts.Presentation;
import org.LexGrid.emf.naming.Mappings;
import org.LexGrid.emf.naming.NamingFactory;
import org.LexGrid.emf.naming.SupportedAssociation;
import org.LexGrid.emf.naming.SupportedCodingScheme;
import org.LexGrid.emf.naming.SupportedHierarchy;
import org.LexGrid.emf.naming.SupportedLanguage;
import org.LexGrid.emf.naming.SupportedProperty;
import org.LexGrid.emf.naming.SupportedPropertyQualifier;
import org.LexGrid.emf.naming.SupportedSource;
import org.LexGrid.emf.relations.Association;
import org.LexGrid.emf.relations.AssociationSource;
import org.LexGrid.emf.relations.AssociationTarget;
import org.LexGrid.emf.relations.Relations;
import org.LexGrid.emf.relations.RelationsFactory;
import org.LexGrid.emf.relations.util.RelationsUtil;
import org.LexGrid.messaging.LgMessageDirectorIF;
import org.LexGrid.messaging.impl.CachingMessageDirectorImpl;
import org.LexGrid.util.sql.DBUtility;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.apache.commons.lang.StringUtils;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import edu.mayo.informatics.lexgrid.convert.utility.loaderPreferences.PreferenceLoaderConstants;

/**
 * This is the Main EMF conversion driver.  The virtually all of the conversion is driven from this
 * class. 
 * 
 * @author <A HREF="mailto:scott.bauer@mayo.edu">Scott Bauer</A>
 * @author <A HREF="mailto:stancl.craig@mayo.edu">Craig Stancl</A>
 * @author <A HREF="mailto:sharma.deepak2@mayo.edu">Deepak Sharma</A>
 *
 */
/**
 * @author <A HREF="mailto:scott.bauer@mayo.edu">Scott Bauer</A>
 * @author <A HREF="mailto:stancl.craig@mayo.edu">Craig Stancl</A>
 * @author <A HREF="mailto:sharma.deepak2@mayo.edu">Deepak Sharma</A>
 * 
 */
public class HL7MapToLexGrid {
    private LgMessageDirectorIF messages_;
    private String accessConnectionString;
    private String driver;
    private String codingScheme;
    private Hashtable conceptsList;
    private Hashtable internalIdConceptCodeMap;
    private LoaderPreferences loaderPrefs;

    public HL7MapToLexGrid(String currentCodingScheme, String database, String driver, LgMessageDirectorIF lg_messages) {

        this.messages_ = new CachingMessageDirectorImpl(lg_messages);
        accessConnectionString = database;
        this.driver = driver;
        this.codingScheme = currentCodingScheme;
        conceptsList = new Hashtable();
        internalIdConceptCodeMap = new Hashtable();
    }

    void initRun(CodingScheme csclass) {

        try {

            loadCodingScheme(csclass, accessConnectionString, driver);
            loadConcepts(csclass, accessConnectionString, driver);

            // Load Metadata if the path has been specified in the Loader
            // Preferences
            if (loaderPrefs != null && loaderPrefs.getXMLMetadataFilePath() != null) {
                loadMetaData(csclass, accessConnectionString, driver);
            } else {
                messages_.info("No metadata file path was specified in the Loader Preferences, not loading Metadata.");
            }

        } catch (Exception e) {
            messages_.error("Failed to connect to RIM Database, check connection values");
            e.printStackTrace();
        } finally {

        }
    }

    void loadCodingScheme(CodingScheme csclass, Connection c) {

        try {

            // TODO Load any persisted HL7 rim metadata as coding scheme data
            // here
            // TODO Get the RIM data elements from the Model Table and persist
            // to the
            // emf csclass object as is done below
            // TODO use the statement below later in an iterative loop to get
            // the
            // top node and possibly persist the metadata for each scheme as
            // properties.
            PreparedStatement getCodingSchemeInfo = c.prepareStatement("SELECT codeSystemid, "
                    + "codeSystemType, codeSystemName, fullName, description, "
                    + "releaseId, copyrightNotice FROM VCS_Code_System WHERE codeSystemid = ?");
            getCodingSchemeInfo.setString(1, codingScheme);
            ResultSet results = getCodingSchemeInfo.executeQuery();
            results.next();

            String name = results.getString("codeSystemName");

            csclass.setCodingSchemeName(name);
            csclass.setCodingSchemeURI(results.getString("codeSystemid"));
            csclass.setFormalName(results.getString("FullName"));
            csclass.setEntityDescription(results.getString("description"));
            csclass.setDefaultLanguage("en");

            String version = results.getString("releaseId");
            if (version != null && version.length() > 0)
                csclass.setRepresentsVersion(version);
            else {
                csclass.setRepresentsVersion(SQLTableConstants.TBLCOLVAL_MISSING);
            }
            Text txt = CommontypesFactory.eINSTANCE.createText();
            txt.setValue((String) results.getString("copyrightNotice"));
            csclass.setCopyright(txt);
            csclass.setMappings((Mappings) NamingFactory.eINSTANCE.createMappings());

            // Add SupportedCodingScheme and SupportedLanguage Mappings
            SupportedCodingScheme scs = NamingFactory.eINSTANCE.createSupportedCodingScheme();
            scs.setLocalId(csclass.getCodingSchemeName());
            scs.setUri(csclass.getCodingSchemeURI());
            csclass.getMappings().getSupportedCodingScheme().add(scs);

            SupportedLanguage lang = NamingFactory.eINSTANCE.createSupportedLanguage();
            lang.setLocalId(HL72EMFConstants.DEFAULT_LANGUAGE_EN);
            csclass.getMappings().getSupportedLanguage().add(lang);

            results.close();
        } catch (Exception e) {
            messages_.error("Failed while preparing HL7 Coding Scheme Class", e);
            e.printStackTrace();
        }
    }

    void loadCodingScheme(CodingScheme csclass, String connectionString, String driver) {
        Connection c = null;
        try {

            c = DBUtility.connectToDatabase(accessConnectionString, driver, null, null);

            PreparedStatement getCodingSchemeInfo = c
                    .prepareStatement("SELECT modelID, name, versionNumber, description FROM Model");
            ResultSet results = getCodingSchemeInfo.executeQuery();
            results.next();
            String name = results.getString("modelID");
            csclass.setCodingSchemeName(name);
            csclass.setCodingSchemeURI(HL72EMFConstants.DEFAULT_URN);
            csclass.setFormalName(results.getString("name"));
            csclass.setEntityDescription(results.getString("description"));
            csclass.setDefaultLanguage(HL72EMFConstants.DEFAULT_LANGUAGE_EN);
            String version = results.getString("versionNumber");
            if (version != null && version.length() > 0)
                csclass.setRepresentsVersion(version);
            else {
                csclass.setRepresentsVersion(SQLTableConstants.TBLCOLVAL_MISSING);
            }
            Text txt = CommontypesFactory.eINSTANCE.createText();
            txt.setValue("copyrightNotice goes here");

            csclass.setCopyright(txt);
            csclass.setMappings((Mappings) NamingFactory.eINSTANCE.createMappings());

            // Add SupportedCodingScheme and SupportedLanguage Mappings
            SupportedCodingScheme scs = NamingFactory.eINSTANCE.createSupportedCodingScheme();
            scs.setLocalId(csclass.getCodingSchemeName());
            scs.setUri(csclass.getCodingSchemeURI());
            csclass.getMappings().getSupportedCodingScheme().add(scs);

            SupportedLanguage lang = NamingFactory.eINSTANCE.createSupportedLanguage();
            lang.setLocalId(HL72EMFConstants.DEFAULT_LANGUAGE_EN);
            csclass.getMappings().getSupportedLanguage().add(lang);
            results.close();
            // Add SupportedHierarchy Mappings
            SupportedHierarchy hierarchy = NamingFactory.eINSTANCE.createSupportedHierarchy();
            hierarchy.setLocalId(HL72EMFConstants.DEFAULT_ASSOC);
            ArrayList list = new ArrayList();
            list.add(HL72EMFConstants.DEFAULT_ASSOC);
            hierarchy.setAssociationNames(list);
            hierarchy.setRootCode(HL72EMFConstants.DEFAULT_ROOT_NODE);
            hierarchy.setIsForwardNavigable(true);
            csclass.getMappings().getSupportedHierarchy().add(hierarchy);

        } catch (Exception e) {
            messages_.error("Failed while preparing HL7 Coding Scheme Class", e);
            e.printStackTrace();
        } finally {
            try {
                c.close();
            } catch (SQLException e) {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
    }

    void loadConcepts(CodingScheme csclass, String connectionString, String driver) {

        Connection c = null;
        String association_name = null;
        Relations relations = null;
        Association parent_assoc = null;
        Association hasSubtypeAssociation = null;

        Entities concepts = csclass.getEntities();
        if (concepts == null) {
            concepts = ConceptsFactory.eINSTANCE.createEntities();
            csclass.setEntities(concepts);
        }
        int num = 0;
        try {
            c = DBUtility.connectToDatabase(accessConnectionString, driver, null, null);
            // Pre-Load the supported associations -- this will just be the
            // hasSubtype association for 99% of the contained coding schemes.
            relations = RelationsFactory.eINSTANCE.createRelations();
            relations.setContainerName("relations");
            csclass.getRelations().add(relations);

            PreparedStatement getAssociations = c
                    .prepareStatement("SELECT distinct(relationCode) FROM VCS_concept_relationship");
            ResultSet associations = getAssociations.executeQuery();

            // This assures that the single other association (smaller_than)
            // will be supported and provides future support for associations
            // other than hasSubtype
            while (associations.next()) {
                association_name = associations.getString("relationCode");
                SupportedAssociation sa = NamingFactory.eINSTANCE.createSupportedAssociation();
                sa.setLocalId(association_name);
                csclass.getMappings().getSupportedAssociation().add(sa);

                parent_assoc = RelationsFactory.eINSTANCE.createAssociation();
                parent_assoc.setEntityCode(association_name);
                parent_assoc.setForwardName(association_name);
                parent_assoc.setIsTransitive(true);
                RelationsUtil.subsume(relations, parent_assoc);
                messages_.info("Loading association: " + association_name);

                // Save a reference to the hasSubtype association so we can
                // use it to attach the artificial top nodes to the "@" node
                if (association_name.equals("hasSubtype")) {
                    hasSubtypeAssociation = parent_assoc;
                }
            }

            associations.close();

            // Pre-load all the supported properties
            PreparedStatement getProperties = c
                    .prepareStatement("SELECT distinct(propertyCode) FROM VCS_concept_property");
            ResultSet properties = getProperties.executeQuery();

            String property = null;
            while (properties.next()) {
                property = properties.getString("propertyCode");
                SupportedProperty sp = NamingFactory.eINSTANCE.createSupportedProperty();
                sp.setLocalId(property);
                if (!csclass.getMappings().getSupportedProperty().contains(sp))
                    csclass.getMappings().getSupportedProperty().add(sp);
            }
            properties.close();

            // Pre-load the supported property qualifier source-code
            String propertyQualifier = "source-code";
            SupportedPropertyQualifier spq = NamingFactory.eINSTANCE.createSupportedPropertyQualifier();
            spq.setLocalId(propertyQualifier);
            if (!csclass.getMappings().getSupportedPropertyQualifier().contains(spq))
                csclass.getMappings().getSupportedPropertyQualifier().add(spq);

            // Pre-load all the supported sources
            PreparedStatement getSources = c.prepareStatement("SELECT distinct(codeSystemName) FROM VCS_code_system");
            ResultSet sources = getSources.executeQuery();

            String source = null;
            while (sources.next()) {
                source = sources.getString("codeSystemName");
                SupportedSource ss = NamingFactory.eINSTANCE.createSupportedSource();

                ss.setLocalId(source);
                if (!csclass.getMappings().getSupportedSource().contains(ss))
                    csclass.getMappings().getSupportedSource().add(ss);
            }

            sources.close();

            // Create the artificial top nodes, the @ node.
            // Persist the top nodes to a list
            // Create a method do handle them as concepts.
            PreparedStatement getTopNodes = c.prepareStatement("SELECT distinct(codeSystemid) FROM VCS_code_system");
            ResultSet topNodes = getTopNodes.executeQuery();
            while (topNodes.next()) {
                loadArtificialTopNodes(csclass, concepts, connectionString, driver, topNodes.getString(1),
                        hasSubtypeAssociation, c);
            }
            topNodes.close();

            // Pre-load some concept data into hash tables.
            PreparedStatement getConcept = c.prepareStatement("SELECT internalId, conceptCode2, "
                    + SQLTableConstants.TBLCOL_CONCEPTSTATUS + " FROM VCS_concept_code_xref");
            ResultSet results = getConcept.executeQuery();
            int i = 0;
            while (results.next()) {
                int internalId = results.getInt("internalId");
                String conceptCode = Integer.toString(internalId) + ":" + results.getString("conceptCode2");
                String status = results.getString(SQLTableConstants.TBLCOL_CONCEPTSTATUS);
                conceptsList.put(new Integer(i), new HL7ConceptContainer(Integer.toString(internalId), conceptCode,
                        status));
                internalIdConceptCodeMap.put(new Integer(internalId), conceptCode);
                i++;
            }
            results.close();
        } catch (Exception e) {
            messages_.error("Failed while preparing HL7 concepts.", e);
            e.printStackTrace();
        } finally {
            try {
                c.close();
            } catch (SQLException e) {
                // TODO drop stack trace.
                e.printStackTrace();
            }
        }

        // process the concept data
        for (int j = 0; j < conceptsList.size(); j++) {
            Concept concept = ConceptsFactory.eINSTANCE.createConcept();
            HL7ConceptContainer conceptContainer = (HL7ConceptContainer) conceptsList.get(new Integer(j));
            String uniqueInternalId = conceptContainer.getInternalId();
            concept.setEntityCode(conceptContainer.getConceptCode());
            loadPresentation(csclass, concept, uniqueInternalId, connectionString, driver);
            loadConceptProperties(csclass, concept, uniqueInternalId, connectionString, driver);
            loadRelations(csclass, concept, uniqueInternalId, parent_assoc, relations, connectionString, driver,
                    internalIdConceptCodeMap);
            concepts.getEntity().add(concept);
            num++;
            messages_.info("Processed Concept code: " + conceptContainer.getConceptCode() + " number of concepts: "
                    + num);
        }
        csclass.setApproxNumConcepts(concepts.getEntity().size());
        messages_.info("Concepts added=" + num);

    }

    void loadArtificialTopNodes(CodingScheme csclass, Entities concepts, String connectionString, String driver,
            String node, Association parent_assoc, Connection c) {

        // Create an "@" top node.
        Concept topNode = ConceptsFactory.eINSTANCE.createConcept();

        // Create and set the concept code
        String topNodeDesignation = "@";
        topNode.setEntityCode(topNodeDesignation);
        topNode.setIsAnonymous(Boolean.TRUE);
        topNode.setEntityDescription("Root node for subclass relations.");

        try {

            PreparedStatement getArtificialTopNodeData = c
                    .prepareStatement("SELECT codeSystemid, codeSystemName, fullName FROM "
                            + "VCS_Code_System WHERE codeSystemid = ?");
            getArtificialTopNodeData.setString(1, node);
            ResultSet results = getArtificialTopNodeData.executeQuery();

            while (results.next()) {
                Concept concept = ConceptsFactory.eINSTANCE.createConcept();

                // Create and set the concept code
                concept.setEntityCode(results.getString("codeSystemName"));
                concept.setEntityDescription(results.getString("fullName"));
                concepts.getEntity().add(concept);
                // Subsume this artificial concept to the topNode "@"
                try {
                    AssociationSource ai = RelationsFactory.eINSTANCE.createAssociationSource();
                    ai.setSourceEntityCode(topNode.getEntityCode());
                    ai = RelationsUtil.subsume(parent_assoc, ai);

                    AssociationTarget at = RelationsFactory.eINSTANCE.createAssociationTarget();
                    at.setTargetEntityCode(concept.getEntityCode());
                    RelationsUtil.subsume(ai, at);
                } catch (Exception e) {
                    messages_.error("Failed while creating artificial top nodes for HL7 RIM", e);
                    e.printStackTrace();
                }

                // Now find the top nodes of the scheme and subsume them to this
                // scheme's artificial top node.
                // First get a list of all nodes in the scheme.
                ArrayList topNodes = new ArrayList();
                String codingSchemeId = results.getString("codeSystemid");
                PreparedStatement getSystemCodes = c
                        .prepareStatement("SELECT DISTINCT (internalId)FROM VCS_concept_code_xref WHERE codeSystemId2 =?");
                getSystemCodes.setString(1, codingSchemeId);
                ResultSet systemCodes = getSystemCodes.executeQuery();
                while (systemCodes.next()) {
                    topNodes.add(systemCodes.getString("internalId"));
                }
                systemCodes.close();

                // Drop any from the list that aren't top nodes.
                ArrayList nodesToRemove = new ArrayList();
                for (int i = 0; i < topNodes.size(); i++) {
                    PreparedStatement checkForTopNode = c
                            .prepareStatement("SELECT targetInternalId FROM VCS_concept_relationship WHERE targetInternalId =?");
                    checkForTopNode.setString(1, (String) topNodes.get(i));
                    ResultSet isTopNode = checkForTopNode.executeQuery();
                    if (isTopNode.next()) {
                        if (topNodes.get(i).equals(isTopNode.getString(1))) {
                            nodesToRemove.add(topNodes.get(i));
                        }
                        checkForTopNode.close();
                    }
                }
                for (int i = 0; i < nodesToRemove.size(); i++) {
                    topNodes.remove(nodesToRemove.get(i));
                }

                // Get the full concept code for each.
                for (int i = 0; i < topNodes.size(); i++) {
                    PreparedStatement getconceptSuffix = c
                            .prepareStatement("SELECT conceptCode2 FROM VCS_concept_code_xref where internalId = ?");

                    getconceptSuffix.setString(1, (String) topNodes.get(i));
                    ResultSet conceptCode = getconceptSuffix.executeQuery();
                    conceptCode.next();
                    topNodes.set(i, topNodes.get(i) + ":" + conceptCode.getString(1));
                }

                // For each top node subsume to the current artificial node for
                // the scheme.
                for (int j = 0; j < topNodes.size(); j++) {
                    try {
                        AssociationSource ai = RelationsFactory.eINSTANCE.createAssociationSource();
                        ai.setSourceEntityCode(concept.getEntityCode());
                        ai = RelationsUtil.subsume(parent_assoc, ai);

                        AssociationTarget at = RelationsFactory.eINSTANCE.createAssociationTarget();
                        at.setTargetEntityCode((String) topNodes.get(j));
                        RelationsUtil.subsume(ai, at);
                    } catch (Exception e) {
                        messages_.error("Failed while processing HL7 psuedo top node hierarchy", e);
                        e.printStackTrace();
                    }
                }

            }

        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

    }

    void loadPresentation(CodingScheme csclass, Concept concept, Connection c, String uniqueInternalId) {

        try {
            // Pull source code system name and designation
            PreparedStatement getEntityDescription = c
                    .prepareStatement("SELECT DISTINCT b.codeSystemId2, a.designation, "
                            + "a.language,a.preferredForLanguage, c.codeSystemName "
                            + " FROM VCS_concept_designation AS a, VCS_concept_code_xref AS b, "
                            + "VCS_code_system AS c " + "WHERE ((a.internalId = ?) AND "
                            + "(a.internalId=b.internalId) AND (c.codeSystemid=b.codeSystemId2))");

            getEntityDescription.setString(1, uniqueInternalId);
            ResultSet edResults = getEntityDescription.executeQuery();

            String entityDescription;
            String sourceCodeSystemName;
            String conceptCodeFull;
            String conceptCode;

            // Determine the source coding scheme concept code
            conceptCodeFull = concept.getEntityCode();
            int colonPosition = conceptCodeFull.lastIndexOf(":");
            conceptCode = conceptCodeFull.substring(colonPosition + 1);

            int presentationCount = 1;
            if (edResults.next()) {
                do {

                    sourceCodeSystemName = edResults.getString("codeSystemName");
                    entityDescription = edResults.getString("designation");

                    Presentation p = ConceptsFactory.eINSTANCE.createPresentation();
                    Text txt = CommontypesFactory.eINSTANCE.createText();
                    txt.setValue((String) entityDescription);
                    p.setValue(txt);

                    if (edResults.getString("preferredForLanguage").equals("1")) {
                        p.setIsPreferred(Boolean.TRUE);
                        concept.setEntityDescription(entityDescription);
                    } else { // Designation is not preferred for language, set
                             // to false
                        p.setIsPreferred(Boolean.FALSE);
                    }
                    p.setPropertyName(HL72EMFConstants.PROPERTY_PRINTNAME);
                    p.setPropertyId("T" + presentationCount);
                    p.setLanguage(edResults.getString("language"));

                    // Set the Qualifier
                    PropertyQualifier propQual = CommontypesFactory.eINSTANCE.createPropertyQualifier();
                    String tag = "source-code";
                    propQual.setPropertyQualifierName(tag);
                    txt = CommontypesFactory.eINSTANCE.createText();
                    txt.setValue((String) conceptCode);
                    propQual.setValue(txt);
                    p.getPropertyQualifier().add(propQual);

                    // Set the Source
                    Source s = CommontypesFactory.eINSTANCE.createSource();
                    s.setValue(sourceCodeSystemName);
                    p.getSource().add(s);

                    concept.getPresentation().add(p);
                    presentationCount++;

                } while (edResults.next());
            } else { // There are no designations, so specify they are missing
                Presentation p = ConceptsFactory.eINSTANCE.createPresentation();
                p.setPropertyName(HL72EMFConstants.PROPERTY_PRINTNAME);
                p.setPropertyId("T" + presentationCount);
                p.setLanguage(HL72EMFConstants.DEFAULT_LANGUAGE_EN);
                concept.getPresentation().add(p);
                entityDescription = HL72EMFConstants.MISSING;
                Text txt = CommontypesFactory.eINSTANCE.createText();
                txt.setValue((String) entityDescription);
                p.setValue(txt);
                p.setIsPreferred(Boolean.TRUE);
                concept.setEntityDescription(entityDescription);

            }
            edResults.close();

            Presentation pd = ConceptsFactory.eINSTANCE.createPresentation();
            PreparedStatement getPresentations = c
                    .prepareStatement("SELECT description, language FROM VCS_concept_description WHERE internalId = ?");
            getPresentations.setString(1, uniqueInternalId);
            ResultSet presResults = getPresentations.executeQuery();

            int definitionCount = 1;
            String description = null;

            if (presResults.next()) {
                // removes HTML tags from description text.
                description = presResults.getString("description");
                description = description.replaceAll("</?[A-Z]+\\b[^>]*>", "");
                description = description.replaceAll("</?[a-z]+\\b[^>]*>", "");
                if (StringUtils.isBlank(description)) {
                    description = HL72EMFConstants.MISSING;
                    messages_.info("Found an empty description on Concept " + conceptCodeFull);
                }

                Text txt = CommontypesFactory.eINSTANCE.createText();
                txt.setValue((String) description);
                pd.setValue(txt);
                pd.setIsPreferred(Boolean.TRUE);
                pd.setPropertyName(HL72EMFConstants.PROPERTY_DEFINITION);
                pd.setLanguage(presResults.getString("language"));

                pd.setPropertyId("D" + definitionCount);
            }

            else {
                Text txt = CommontypesFactory.eINSTANCE.createText();
                txt.setValue((String) HL72EMFConstants.MISSING);
                pd.setValue(txt);
                pd.setIsPreferred(Boolean.TRUE);
                pd.setPropertyName(HL72EMFConstants.PROPERTY_DEFINITION);
                pd.setLanguage(HL72EMFConstants.DEFAULT_LANGUAGE_EN);

                pd.setPropertyId("D" + definitionCount);

            }
            concept.getPresentation().add(pd);
            presResults.close();

        } catch (Exception e) {
            e.printStackTrace();
            messages_.error("Failed to load HL7 Concept presentation", e);

        }

    }

    void loadPresentation(CodingScheme csclass, Concept concept, String uniqueInternalId, String connectionString,
            String driver) {

        Connection c = null;
        try {
            c = DBUtility.connectToDatabase(accessConnectionString, driver, null, null);
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        loadPresentation(csclass, concept, c, uniqueInternalId);
        try {
            c.close();
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    void loadConceptProperties(CodingScheme csclass, Concept concept, Connection c, String uniqueInternalId) {
        try {

            // TODO load this up front with the Relations We may need to do the
            // same with the presentation.
            PreparedStatement getProperties = c
                    .prepareStatement("SELECT propertyCode, propertyValue, language FROM VCS_concept_property WHERE internalId = ?");
            getProperties.setString(1, uniqueInternalId);
            ResultSet properties = getProperties.executeQuery();

            while (properties.next()) {
                Property cp = CommontypesFactory.eINSTANCE.createProperty();
                String propertyLanguage = properties.getString("language");
                if (propertyLanguage == null || propertyLanguage.length() < 1) {
                    cp.setLanguage(HL72EMFConstants.DEFAULT_LANGUAGE_EN);
                } else {
                    cp.setLanguage(propertyLanguage);
                }
                String property = properties.getString("propertyCode");
                cp.setPropertyName(property);
                cp.setPropertyId("P" + concept.getProperty().size());
                Text txt = CommontypesFactory.eINSTANCE.createText();
                txt.setValue((String) properties.getString("propertyValue"));
                cp.setValue(txt);
                concept.getProperty().add(cp);
            }
            properties.close();

        } catch (Exception e) {
            messages_.error("Problem processing concept properties", e);
            e.printStackTrace();
        }

    }

    void loadConceptProperties(CodingScheme csclass, Concept concept, String uniqueInternalId, String connectionString,
            String driver) {
        Connection c = null;
        try {
            c = DBUtility.connectToDatabase(accessConnectionString, driver, null, null);
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        loadConceptProperties(csclass, concept, c, uniqueInternalId);
        try {
            c.close();
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    void loadRelations(CodingScheme csclass, Concept concept, Connection c, String uniqueInternalId,
            Association parent_assoc, Relations relations, Hashtable concept2Id) {
        Hashtable associationsList = new Hashtable();
        try {
            PreparedStatement getSourceAssociations = c
                    .prepareStatement("SELECT relationCode, sourceInternalId, targetInternalId FROM VCS_concept_relationship where sourceInternalId = ?");
            getSourceAssociations.setString(1, uniqueInternalId);
            ResultSet associations = getSourceAssociations.executeQuery();
            int i = 0;
            while (associations.next()) {
                associationsList.put(new Integer(i), new HL7AssocContainer(associations.getString("relationCode"),
                        associations.getInt("sourceInternalId"), associations.getInt("targetInternalId")));
                i++;
            }
            associations.close();
        } catch (Exception e) {
            messages_.error("Failed while getting HL7 relations from RIM database", e);
            e.printStackTrace();
        }

        for (int j = 0; j < associationsList.size(); j++) {
            HL7AssocContainer assoContainer = (HL7AssocContainer) associationsList.get(new Integer(j));
            int sourceCode = assoContainer.getSourceCode();
            int targetCode = assoContainer.getTargetCode();
            Association parent_association = (Association) RelationsUtil.resolveAssociations(csclass,
                    assoContainer.getAssociation()).get(0);

            // TODO enclose entire association setup in some kind of try catch.
            try {
                AssociationSource ai = RelationsFactory.eINSTANCE.createAssociationSource();
                ai.setSourceEntityCode((String) concept2Id.get(new Integer(sourceCode)));
                ai = RelationsUtil.subsume(parent_association, ai);

                AssociationTarget at = RelationsFactory.eINSTANCE.createAssociationTarget();
                at.setTargetEntityCode((String) concept2Id.get(new Integer(targetCode)));
                RelationsUtil.subsume(ai, at);
            } catch (Exception e) {
                messages_.error("Failed while processing HL7 association hierarchy", e);
                e.printStackTrace();
            }

        }
    }

    void loadRelations(CodingScheme csclass, Concept concept, String uniqueInternalId, Association association,
            Relations relations, String connectionString, String driver, Hashtable concept2Id) {
        Connection c = null;
        try {
            c = DBUtility.connectToDatabase(accessConnectionString, driver, null, null);
        } catch (Exception e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        loadRelations(csclass, concept, c, uniqueInternalId, association, relations, concept2Id);
        try {
            c.close();
        } catch (SQLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    void loadMetaData(CodingScheme csclass, String connectionString, String driver) {

        Connection c = null;

        // create the filename of the metadata file to be created
        String fileName = PreferenceLoaderConstants.META_HL7_METADATA_FILE_NAME;

        // String filename = getDBLocation() + "_metadata.xml";
        String filename = loaderPrefs.getXMLMetadataFilePath() + "/" + fileName;

        FileOutputStream fos;
        try {
            fos = new FileOutputStream(filename);

            OutputFormat of = new OutputFormat("XML", "ISO-8859-1", true);
            of.setIndent(1);
            of.setIndenting(true);
            XMLSerializer serializer = new XMLSerializer(fos, of);

            // SAX2.0 ContentHandler.
            ContentHandler hd = serializer.asContentHandler();
            hd.startDocument();

            // Element attributes.
            AttributesImpl atts = new AttributesImpl();

            // CODINGSCHEMES tag.
            hd.startElement("", "", "codingSchemes", atts);

            String defaultLanguage = "en";
            String isNative = "0";
            String dataMissing = SQLTableConstants.TBLCOLVAL_MISSING;
            String codeSystemId;
            String codeSystemType;
            String codeSystemName;
            String fullName;
            String description;
            String releaseId;
            String copyrightNotice;
            Integer approximateNumberofConcepts;

            try {
                c = DBUtility.connectToDatabase(accessConnectionString, driver, null, null);

                PreparedStatement getCodingSchemeMetaData = c
                        .prepareStatement("SELECT b.codeSystemid, b.codeSystemType, b.codeSystemName, "
                                + "b.fullName, b.description, b.releaseId, b.copyrightNotice "
                                + "FROM VCS_code_system AS b");
                ResultSet codingSchemeMetaData = getCodingSchemeMetaData.executeQuery();

                PreparedStatement getCodingSchemeConceptCount = c
                        .prepareStatement("SELECT VCS_concept_code_xref.codeSystemId2, "
                                + "COUNT (VCS_concept_code_xref.codeSystemId2) as conceptcount "
                                + "FROM VCS_concept_code_xref " + "GROUP BY VCS_concept_code_xref.codeSystemId2;");

                ResultSet codingSchemeConceptCount = getCodingSchemeConceptCount.executeQuery();

                Hashtable conceptCountList = new Hashtable();

                while (codingSchemeConceptCount.next()) {

                    String codingSchemeId = codingSchemeConceptCount.getString("codeSystemId2");
                    String codingSchemeCount = codingSchemeConceptCount.getString("conceptcount");

                    conceptCountList.put(new String(codingSchemeId), new Integer(codingSchemeCount));
                }

                codingSchemeConceptCount.close();

                int codeSchemeCounter = 0;

                while (codingSchemeMetaData.next()) {
                    codeSystemId = codingSchemeMetaData.getString("codeSystemid");
                    if (codeSystemId == null) {
                        codeSystemId = SQLTableConstants.TBLCOLVAL_MISSING;
                    }

                    codeSystemType = codingSchemeMetaData.getString("codeSystemType");
                    if (codeSystemType == null) {
                        codeSystemType = SQLTableConstants.TBLCOLVAL_MISSING;
                    }

                    codeSystemName = codingSchemeMetaData.getString("codeSystemName");
                    if (codeSystemName == null) {
                        codeSystemName = SQLTableConstants.TBLCOLVAL_MISSING;
                    }

                    fullName = codingSchemeMetaData.getString("fullName");
                    if (fullName == null) {
                        fullName = SQLTableConstants.TBLCOLVAL_MISSING;
                    }

                    description = codingSchemeMetaData.getString("description");
                    if (description == null) {
                        description = SQLTableConstants.TBLCOLVAL_MISSING;
                    }
                    // The description contains HTML tags. We try to remove
                    // them.
                    int begin = description.lastIndexOf("<p>");
                    int end = description.lastIndexOf("</p>");
                    if (begin > -1) {
                        description = description.substring(begin + 3, end);
                    }
                    // description =
                    // description.replaceAll("</?[A-Z]+\\b[^>]*>", ""); // CRS
                    // description =
                    // description.replaceAll("</?[a-z]+\\b[^>]*>", ""); // CRS
                    // description = description.replaceAll("&#xd;", "");

                    releaseId = codingSchemeMetaData.getString("releaseId");
                    if (releaseId == null) {
                        releaseId = SQLTableConstants.TBLCOLVAL_MISSING;
                    }

                    copyrightNotice = codingSchemeMetaData.getString("copyrightNotice");
                    if (copyrightNotice == null) {
                        copyrightNotice = SQLTableConstants.TBLCOLVAL_MISSING;
                    }

                    approximateNumberofConcepts = ((Integer) conceptCountList.get(codeSystemId));
                    if (approximateNumberofConcepts == null) {
                        approximateNumberofConcepts = new Integer(0);
                    }

                    // Begin codingScheme element
                    atts.clear();
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_CODINGSCHEMENAME, "CDATA", codeSystemName); // May
                                                                                                               // want
                                                                                                               // to
                                                                                                               // change
                                                                                                               // to
                                                                                                               // _fullName
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_FORMALNAME, "CDATA", fullName);
                    // atts.addAttribute("","",SQLTableConstants.TBLCOL_REGISTEREDNAME,"CDATA",
                    // codeSystemId);
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_CODINGSCHEMEURI, "CDATA", codeSystemId);
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_DEFAULTLANGUAGE, "CDATA", defaultLanguage);
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_REPRESENTSVERSION, "CDATA", releaseId);
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_ISNATIVE, "CDATA", isNative);
                    atts.addAttribute("", "", SQLTableConstants.TBLCOL_APPROXNUMCONCEPTS, "CDATA",
                            approximateNumberofConcepts.toString());
                    // atts.addAttribute("","",SQLTableConstants.TBLCOL_FIRSTRELEASE,"CDATA",dataMissing);
                    // atts.addAttribute("","",SQLTableConstants.TBLCOL_MODIFIEDINRELEASE,"CDATA",dataMissing);
                    // atts.addAttribute("","",SQLTableConstants.TBLCOL_DEPRECATED,"CDATA",
                    // dataMissing);
                    hd.startElement("", "", SQLTableConstants.TBLCOL_CODINGSCHEME, atts);

                    // localname
                    atts.clear();
                    hd.startElement("", "", "localName", atts);
                    hd.characters(codeSystemName.toCharArray(), 0, codeSystemName.length());
                    hd.endElement("", "", SQLTableConstants.TBLCOLVAL_LOCALNAME);

                    // entityDescription
                    atts.clear();
                    hd.startElement("", "", SQLTableConstants.TBLCOL_ENTITYDESCRIPTION, atts);
                    hd.characters(description.toCharArray(), 0, description.length());
                    hd.endElement("", "", SQLTableConstants.TBLCOL_ENTITYDESCRIPTION);

                    // copyright
                    atts.clear();
                    hd.startElement("", "", SQLTableConstants.TBLCOL_COPYRIGHT, atts);
                    hd.characters(copyrightNotice.toCharArray(), 0, copyrightNotice.length());
                    hd.endElement("", "", SQLTableConstants.TBLCOL_COPYRIGHT);

                    // May need to include as Property
                    // // CodingScheme
                    // atts.clear();
                    // hd.startElement("","","CodingScheme",atts);
                    // hd.characters(_codeSystemType.toCharArray(),0,_codeSystemType.length());
                    // hd.endElement("","","CodingScheme");

                    // End codingScheme element
                    hd.endElement("", "", SQLTableConstants.TBLCOL_CODINGSCHEME);

                    codeSchemeCounter++;

                } // End while there are result rows to process

                getCodingSchemeConceptCount.close();

                hd.endElement("", "", "codingSchemes");
                hd.endDocument();
                fos.close();

            } catch (Exception e) {
                messages_.error("Failed while preparing HL7 Code System MetaData.", e);
                e.printStackTrace();
            } finally {
                try {
                    c.close();
                } catch (SQLException e) {
                    // TODO drop stack trace.
                    e.printStackTrace();
                }
            }

            // this is the file not found exception
        } catch (FileNotFoundException e1) {
            // TODO drop stack trace.
            e1.printStackTrace();
        } catch (IOException e) {
            // TODO drop stack trace.
            e.printStackTrace();
        } catch (SAXException e) {
            // TODO drop stack trace.
            e.printStackTrace();
        }

    }

    private String getDBLocation() {

        int dbqLocation = accessConnectionString.lastIndexOf("DBQ");
        String dbFullFilename = accessConnectionString.substring(dbqLocation + 4);
        int sep = dbFullFilename.lastIndexOf(".");
        // strip off the filename extension
        String dbfullPath = dbFullFilename.substring(0, sep);

        return dbfullPath;
    }

    public void setLoaderPrefs(LoaderPreferences loaderPrefs) {
        this.loaderPrefs = loaderPrefs;
    }

}