/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.valuedomain;

import java.io.InputStream;
import java.io.Serializable;
import java.net.URI;

import org.LexGrid.LexBIG.DataModel.Collections.AbsoluteCodingSchemeVersionReferenceList;
import org.LexGrid.LexBIG.DataModel.Core.AbsoluteCodingSchemeVersionReference;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag;
import org.LexGrid.LexBIG.DataModel.Core.LogEntry;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.emf.valueDomains.ValueDomainDefinition;
import org.LexGrid.managedobj.RemoveException;
import org.lexgrid.valuedomain.dto.ResolvedValueDomainCodedNodeSet;
import org.lexgrid.valuedomain.dto.ResolvedValueDomainDefinition;

/**
 * Value Domain extension for LexGrid.
 * 
 * @author <A HREF="mailto:dwarkanath.sridhar@mayo.edu">Sridhar Dwarkanath</A>
 */
public interface LexEVSValueDomainServices extends Serializable {
	
	
	/**
	 * Loads supplied valueDomainDefinition object
	 * @param vddef value domain to load
	 * @param systemReleaseURI
	 * @throws LBException
	 */
	public void loadValueDomain(ValueDomainDefinition vddef, String systemReleaseURI) throws LBException;
	
	/**
	 * Loads value domain using inputStream
	 * @param inputStream
	 * @param failOnAllErrors
	 * @throws Exception
	 */
	public void loadValueDomain(InputStream inputStream, boolean failOnAllErrors) throws Exception;
	
	/**
	 * Loads value domain by reading XML file location supplied
	 * @param xmlFileLocation XML file containing valueDomain definitions
	 * @param failOnAllErrors
	 * @throws Exception
	 */
	public void loadValueDomain(String xmlFileLocation, boolean failOnAllErrors) throws Exception;
	
	/**
	 * Perform validation of the candidate resource without loading data.  
	 * @param uri XML file containing valueDomain definitions
	 * @param validationLevel validate &lt;int&gt; 
	 *         Supported levels of validation include:
	 *         0 = Verify document is well-formed
	 *         1 = Verify document is valid
	 * @throws LBParameterException
	 */
	public void validate(URI uri, int valicationLevel) throws LBParameterException;
	
	/**
	 * Determine if the supplied entity code is a valid result for the supplied
	 * value domain and, if it is, return the particular codingSchemeVersion
	 * that was used.
	 * 
	 * @param entityCode
	 * @param valueDomainURI
	 * @return coding scheme and version if the entityCode is valid, null
	 *         otherwise
	 * @throws LBException
	 */
	public AbsoluteCodingSchemeVersionReference isConceptInDomain(String entityCode, URI valueDomainURI)  throws LBException;

	/**
	 * Determine if the supplied entity code is a valid result for the
	 * supplied domain.
	 * 
	 * @param entityCode
	 *            entity code
	 * @param entityCodeNamespace
	 *            entity code namespace (can be null if the same as the coding
	 *            scheme namespace)
	 * @param csvt
	 *            coding scheme version or tag
	 * @param valueDomainURI
	 * 			  value domain URI
	 * @return The codingScheme version of that asserts that the code is in the domain or null if
	 *         it isn't.
	 * @throws LBException
	 */
	public AbsoluteCodingSchemeVersionReference isConceptInDomain(String entityCode,
			URI entityCodeNamespace, CodingSchemeVersionOrTag csvt,
			URI valueDomainURI) throws LBException;

	/**
	 * Resolve a value domain using the supplied set of coding scheme versions.
	 * 
	 * @param valueDomainURI
	 * 			  value domain URI
	 * @param csVersionList
	 *            list of coding scheme versions to use in resolution. IF the
	 *            value domain uses a version that isn't mentioned in this list,
	 *            the resolve function will return the codingScheme and version 
	 *            that was used as a default for the resolution. 
	 * @return Resolved Value Domain Definition
	 * @throws LBException
	 */
	public ResolvedValueDomainDefinition resolveValueDomain(URI valueDomainURI,
			AbsoluteCodingSchemeVersionReferenceList csVersionList) throws LBException;

	/**
	 * Check whether childValueDomainURI is a child of parentValueDomainURI.
	 * 
	 * @param childValueDomainURI
	 * 			child value domain URI
	 * @param parentValueDomainURI
	 * 			parent value domain URI
	 * @return YES, if all the elements of the child domain are in the parent domain 
	 *  		NO otherwise.
	 *  @throws LBException
	 */
	public boolean isSubDomain(URI childValueDomainURI, URI parentValueDomainURI) throws LBException;

	/**
	 * Returns value domain definition for supplied value domain URI.
	 * 
	 * @param valueDomainURI
	 * 			value domain URI
	 * @return value domain definition
	 * @throws LBException
	 */
	public ValueDomainDefinition getValueDomainDefinition(URI valueDomainURI) throws LBException;	
	
	/**
	 * Return the URI's for the value domain definition(s) for the supplied
	 * domain name. If the name is null, returns everything. If the name is not
	 * null, returns the value domain(s) that have the assigned name. 
	 * 
	 * Note: plural because there is no guarantee of valueDomain uniqueness. If the name is the
	 * empty string "", returns all unnamed valueDomains.
	 * 
	 * @param valueDomainName
	 * @return value domain URI's
	 * @throws LBException
	 */
	public URI[] listValueDomains(String valueDomainName) throws LBException;

	/**
	 * Return the URI's of all unnamed value domain definition(s).
	 * 
	 * @return value domain URI's
	 * @throws LBException
	 */
	public URI[] getAllValueDomainsWithNoNames()  throws LBException;
	
	
	/**
	 * Resolves the value domain supplied and restricts to the term and matchAlgorith supplied.
	 * Return object ResolvedValueDomainCodedNodeSet contains the codingScheme URI and Version that was
	 * used to resolve and the CodedNodeSet.
	 * Note : the CodedNodeSet is unresolved
	 *  
	 * @param term text to search
	 * @param valueDomainURI URI of ValueDomain
	 * @param matchAlgorithm
	 * @return ResolvedValueDomainCodedNodeSet containing the codingScheme URI and Version that was
	 * used to resolve and the CodedNodeSet.
	 * @throws LBException
	 */
	public ResolvedValueDomainCodedNodeSet getValueDomainEntitiesForTerm(String term,
			URI valueDomainURI, String matchAlgorithm) throws LBException;

	
	
	/**
	 * Returns list of coding scheme summary that is referenced by the supplied
	 * value domain. 
	 * @param valueDomainURI
	 * @return coding scheme version reference list
	 * 
	 * @throws LBException
	 */
	public AbsoluteCodingSchemeVersionReferenceList getCodingSchemesInValueDomain(URI valueDomainURI) throws LBException;
	
	/**
	 * Determine if the supplied entity code is of type valueDomain in supplied
	 * coding scheme and, if it is, return the true, otherwise return false.
	 * 
	 * @param entityCode
	 * @param codingSchemeName
	 * @param csvt
	 * @return TRUE : If entityCode is of type valueDomain in supplied coding scheme, FALSE : otherwise
	 * 
	 * @throws LBException
	 */
	public boolean isDomain(String entityCode, String codingSchemeName, CodingSchemeVersionOrTag csvt) throws LBException;
	
	/**
	 * Removes supplied value domain definition from the system.
	 * 
	 * @param valueDomainURI URI of value domain to remove
	 * @throws LBException
	 * @throws RemoveException
	 */
	public void removeValueDomain(URI valueDomainURI) throws LBException, RemoveException ;
	
	/**
	 * Removes all value domain definitions from the system.
	 * 
	 * @throws LBException
	 * @throws RemoveException
	 */
	public void removeAllValueDomains() throws LBException, RemoveException ;
	
	/**
	 * Drops value domain tables only if there are no value domain and pick list entries.
	 * @throws LBException
	 * @throws RemoveException
	 */
	public void dropValueDomainTables() throws LBException, RemoveException ;
	
	public LogEntry[] getLogEntries();
}