/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.valuedomain.impl;

import java.io.File;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.sql.SQLException;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import org.LexGrid.LexBIG.DataModel.Collections.AbsoluteCodingSchemeVersionReferenceList;
import org.LexGrid.LexBIG.DataModel.Collections.LocalNameList;
import org.LexGrid.LexBIG.DataModel.Core.AbsoluteCodingSchemeVersionReference;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag;
import org.LexGrid.LexBIG.DataModel.Core.LogEntry;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.DataModel.InterfaceElements.LoadStatus;
import org.LexGrid.LexBIG.DataModel.InterfaceElements.types.ProcessState;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBInvocationException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.Impl.LexBIGServiceImpl;
import org.LexGrid.LexBIG.Impl.dataAccess.ResourceManager;
import org.LexGrid.LexBIG.Impl.dataAccess.SystemVariables;
import org.LexGrid.LexBIG.Impl.loaders.MessageDirector;
import org.LexGrid.LexBIG.Impl.logging.LgLoggerIF;
import org.LexGrid.LexBIG.Impl.logging.LoggerFactory;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.LexBIG.Utility.Constructors;
import org.LexGrid.LexBIG.Utility.Iterators.ResolvedConceptReferencesIterator;
import org.LexGrid.LexBIG.Utility.LBConstants.MatchAlgorithms;
import org.LexGrid.annotations.LgClientSideSafe;
import org.LexGrid.emf.valueDomains.CodingSchemeReference;
import org.LexGrid.emf.valueDomains.DefinitionEntry;
import org.LexGrid.emf.valueDomains.ValueDomainDefinition;
import org.LexGrid.emf.versions.SystemRelease;
import org.LexGrid.emf.versions.impl.SystemReleaseImpl;
import org.LexGrid.managedobj.FindException;
import org.LexGrid.managedobj.InsertException;
import org.LexGrid.managedobj.ObjectAlreadyExistsException;
import org.LexGrid.managedobj.RemoveException;
import org.LexGrid.managedobj.ServiceInitException;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.LexGrid.valueDomains.types.DefinitionOperator;
import org.apache.commons.lang.StringUtils;
import org.lexgrid.valuedomain.LexEVSValueDomainServices;
import org.lexgrid.valuedomain.dto.EntityReferenceWrapper;
import org.lexgrid.valuedomain.dto.ResolvedValueDomainCodedNodeSet;
import org.lexgrid.valuedomain.dto.ResolvedValueDomainDefinition;
import org.lexgrid.valuedomain.persistence.SystemReleaseServices;
import org.lexgrid.valuedomain.persistence.VDMappingServices;
import org.lexgrid.valuedomain.persistence.VDSServices;
import org.lexgrid.valuedomain.persistence.VDServiceHelper;
import org.lexgrid.valuedomain.persistence.VDServices;
import org.lexgrid.valuedomain.persistence.VDXMLread;

/**
 * Implementation of Value Domain extension for LexGrid.
 * 
 * @author <A HREF="mailto:dwarkanath.sridhar@mayo.edu">Sridhar Dwarkanath</A>
 */
public class LexEVSValueDomainServicesImpl implements LexEVSValueDomainServices {

	// Associated service ...
	private LexBIGService lbs_;
	private VDSServices vdss_;
	private VDServices vds_;
	private VDMappingServices mapS_;
	private VDServiceHelper sh_;
	protected MessageDirector md_;
	protected LoadStatus status_;
	private static final String name_ = "LexEVSValueDomainServicesImpl";
	private static final String desc_ = "Implements LexGrid Value Domain services.";
	private static final String provider_ = "Mayo Clinic";
	private static final String version_ = "1.0";
	/**
	 * 
	 */
	private static final long serialVersionUID = 4995582014921448463L;

	public LexEVSValueDomainServicesImpl() {
		getStatus().setState(ProcessState.PROCESSING);
		getStatus().setStartTime(new Date(System.currentTimeMillis()));
		md_ = new MessageDirector(getName(), status_);
	}

	private LoadStatus getStatus() {
		if (status_ == null)
			status_ = new LoadStatus();
		return status_;
	}

	private LgLoggerIF getLogger() {
		return LoggerFactory.getLogger();
	}
	
	public LogEntry[] getLogEntries(){
		if (md_ != null)
			return md_.getLogEntries(null);
		
		return null;
	}
	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#loadValueDomain(org.LexGrid.emf.valueDomains.ValueDomainDefinition)
	 */
	public void loadValueDomain(ValueDomainDefinition vddef, String systemReleaseURI)
			throws LBException {
		getLogger().logMethod(new Object[] { vddef });
		try {
			if (vddef != null)
			{
				md_.info("Loading value domain definition : " + vddef.getValueDomainURI());				
				getVDService().insert(vddef, systemReleaseURI);
			}
		} catch (ObjectAlreadyExistsException e) {
			md_.fatal("Failed while while loading value domain definition", e);
			throw new LBException("Failed while while loading value domain definition", e);
		} catch (InsertException e) {
			md_.fatal("Failed while while loading value domain definition", e);
			throw new LBException("Failed while while loading value domain definition", e);
		} catch (ServiceInitException e) {
			md_.fatal("Failed while while loading value domain definition", e);
			throw new LBException("Failed while while loading value domain definition", e);
		}
		md_.info("Finished loading value domain definition");
	}

	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#loadValueDomain(java.io.InputStream, boolean)
	 */
	public void loadValueDomain(InputStream inputStream, boolean failOnAllErrors)
			throws LBException {
		getLogger().logMethod(new Object[] { inputStream });
		VDXMLread vdXML = new VDXMLread(inputStream, md_, failOnAllErrors);

		internalLoadVD(vdXML);
	}

	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#loadValueDomain(java.lang.String, boolean)
	 */
	public void loadValueDomain(String xmlFileLocation, boolean failOnAllErrors)
			throws LBException {
		getLogger().logMethod(new Object[] { xmlFileLocation });
		VDXMLread vdXML;
		try {
			vdXML = new VDXMLread(getStringFromURI(new URI(xmlFileLocation)), null, md_,failOnAllErrors);
		} catch (URISyntaxException e) {
			throw new LBException("Failed loading XML.", e);
		}
		
		internalLoadVD(vdXML);
	}

	/**
	 * Common method that will be used for loading value domain to database.
	 * @param vdXML
	 * @throws Exception
	 */
	private void internalLoadVD(VDXMLread vdXML) throws LBException {
		getLogger().logMethod(new Object[] { vdXML });
		
		try 
		{
			ValueDomainDefinition[] vdDefs = vdXML.readAllValueDomains();
			SystemRelease systemRelease = vdXML.getSystemRelease();
			
			String releaseURI = null;
			
			if (systemRelease != null)
			{
				releaseURI = systemRelease.getReleaseURI();
				
				md_.info("Loading System Release : " + releaseURI);
				
				SystemReleaseServices releaseService = (SystemReleaseServices) getVDService().getNestedService(SystemReleaseImpl.class);
				releaseService.insert(systemRelease);
				
				md_.info("Finished Loading System Release : " + releaseURI);
			}
			
			md_.info("Loading Value Domain Definitions");
			for (ValueDomainDefinition vddef : vdDefs) {
				loadValueDomain(vddef, releaseURI);			
			}
			md_.info("Finished Loading Value Domain Definitions");
			
			md_.info("Loading Mappings");
			mapS_ = (VDMappingServices) getVDSService().getNestedService(
					VDMappingServices.class);
			mapS_.insert(vdXML.getVdMappings(), -1);
			md_.info("Finished Loading Mappings");
			
			md_.info("Load Process Complete.");
		} 
		catch (ServiceInitException e)
		{
			throw new LBException("Problem loading valueDomain." + e.getStackTrace());
		} 
		catch (ObjectAlreadyExistsException e) 
		{
			throw new LBException("Problem loading valueDomain." + e.getStackTrace());
		} 
		catch (InsertException e) 
		{
			throw new LBException("Problem loading valueDomain." + e.getStackTrace());
		} 
		catch (Exception e) 
		{
			throw new LBException("Problem loading valueDomain." + e.getStackTrace());
		}
	}
	
	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#validate(java.net.URI, int)
	 */
	public void validate(URI uri, int v1) throws LBParameterException{
		VDXMLread vdXML = new VDXMLread(uri.toString(), null, md_, true);
		vdXML.validate(uri, v1);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.lexgrid.extension.valuedomain.LexEVSValueDomainServices#
	 * getAllValueDomainsWithNoNames()
	 */
	public URI[] getAllValueDomainsWithNoNames() throws LBException {
		getLogger().logMethod(new Object[]{});
		URI[] uris;
		try {
			uris = getServiceHelper().getValueDomainServices().findByValueDomainName(" ");
		} catch (FindException e) {
			md_.fatal("Failed while finding Value Domain Definition with no names", e);
			throw new LBException("Failed while finding Value Domain Definition with no names", e);
		}
		return uris;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.lexgrid.extension.valuedomain.LexEVSValueDomainServices#
	 * getCodingSchemesInValueDomain(java.net.URI)
	 */
	@SuppressWarnings("unchecked")
	public AbsoluteCodingSchemeVersionReferenceList getCodingSchemesInValueDomain(
			URI valueDomainURI) throws LBException {
		getLogger().logMethod(new Object[] { valueDomainURI });
		AbsoluteCodingSchemeVersionReferenceList csList = new AbsoluteCodingSchemeVersionReferenceList();
		
		// Get valueDomain definition object for supplied uri.
		ValueDomainDefinition vd = getServiceHelper().getValueDomain(valueDomainURI);
		
		// get default coding scheme for this value domain and add it to the list
		String defaultCS = vd.getDefaultCodingScheme();
		if (StringUtils.isNotEmpty(defaultCS)) {
			csList = getServiceHelper().getAbsoluteCodingSchemeVersionReference(defaultCS, null);
		}
		
		// iterate through all the definitionEntries and gather codingSchemeReferences and add them to the list 
		Iterator<DefinitionEntry> deIter = vd.getDefinitionEntry().iterator();
		while (deIter.hasNext()) {
			DefinitionEntry de = deIter.next();
			CodingSchemeReference csRef = de.getCodingSchemeReference();
			if (csRef != null) {
				String cs = csRef.getCodingScheme();
				if (StringUtils.isNotEmpty(cs) && !cs.equals(defaultCS)) {
					Iterator<AbsoluteCodingSchemeVersionReference> aList = getServiceHelper().getAbsoluteCodingSchemeVersionReference(cs, null).iterateAbsoluteCodingSchemeVersionReference();
					
					while(aList.hasNext()){
						csList.addAbsoluteCodingSchemeVersionReference(aList.next());
					}
				}
			}
		}
		return csList;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.lexgrid.extension.valuedomain.LexEVSValueDomainServices#
	 * getValueDomainDefinition (java.net.URI)
	 */
	public ValueDomainDefinition getValueDomainDefinition(URI valueDomainURI)
			throws LBException {
		getLogger().logMethod(new Object[] { valueDomainURI });
		return getServiceHelper().getValueDomain(valueDomainURI);
	}

	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#getValueDomainEntitiesForTerm(java.lang.String, java.net.URI, java.lang.String)
	 */
	public ResolvedValueDomainCodedNodeSet getValueDomainEntitiesForTerm(
			String term, URI valueDomainURI, String matchAlgorithm)
			throws LBException {
		getLogger().logMethod(new Object[] { term, valueDomainURI, matchAlgorithm });
		ResolvedValueDomainCodedNodeSet rvdcns = null;
		try {
			
			// populate valueDomainDefinition object for supplied uri.
			ValueDomainDefinition vdDef = getServiceHelper().getValueDomain(valueDomainURI);
			
			if (vdDef == null)
			{
				md_.fatal("Value domain uri : '"+ valueDomainURI + "' not found.");
				throw new LBException("Value domain uri : '"+ valueDomainURI + "' not found.");
			}
			
			// populate resolvedValueDomainCodedNodeSet
			rvdcns = getServiceHelper().getCodedNodeSetForValueDomain(vdDef, null, null);
			
			// apply the term restriction to the codedNodeSet 
			if (rvdcns != null)
			{
				AbsoluteCodingSchemeVersionReferenceList csvList = rvdcns.getCodingSchemeVersionRefList();
				String csName = null;
				String tag = null;
				String version = null;
				
				if (csvList != null && csvList.getAbsoluteCodingSchemeVersionReferenceCount() == 1)
				{
					AbsoluteCodingSchemeVersionReference csv = csvList.getAbsoluteCodingSchemeVersionReference(0);
					csName = csv.getCodingSchemeURN();
					version = csv.getCodingSchemeVersion();
					CodedNodeSet cns = getServiceHelper().getLexBIGService().getCodingSchemeConcepts(csName, Constructors.createCodingSchemeVersionOrTag(tag, version));
					cns.restrictToMatchingDesignations(term, null, matchAlgorithm, null);
					rvdcns.getCodedNodeSet().intersect(cns);
				}
				else
				{
					rvdcns.getCodedNodeSet().restrictToMatchingDesignations(term, null, matchAlgorithm == null ? MatchAlgorithms.LuceneQuery.name() : matchAlgorithm, null);
				}
			}
			
		} catch (FindException e) {
			md_.fatal("Failed while getting Value Domain entries for term '" + term +"'", e);
			throw new LBException("Failed while getting Value Domain entries for term '" + term +"'", e);
		} catch (ServiceInitException e) {
			md_.fatal("Failed while getting Value Domain entries for term '" + term +"'", e);
			throw new LBException("Failed while getting Value Domain entries for term '" + term +"'", e);
		} catch (URISyntaxException e) {
			md_.fatal("Failed while getting Value Domain entries for term '" + term +"'", e);
			throw new LBException("Failed while getting Value Domain entries for term '" + term +"'", e);
		}
		return rvdcns;
	}
	
	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#isConceptInDomain(java.lang.String, java.net.URI)
	 */
	public AbsoluteCodingSchemeVersionReference isConceptInDomain(
			String entityCode, URI valueDomainURI) throws LBException {
		getLogger().logMethod(new Object[] { entityCode, valueDomainURI });
		return isConceptInDomain(entityCode, null, null, valueDomainURI);
	}

	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#isConceptInDomain(java.lang.String, java.net.URI, org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag, java.net.URI)
	 */
	public AbsoluteCodingSchemeVersionReference isConceptInDomain(String entityCode, URI entityCodeNamespace,
			CodingSchemeVersionOrTag csvt, URI valueDomainURI)throws LBException {
		getLogger().logMethod(new Object[] { entityCode, entityCodeNamespace, csvt, valueDomainURI });
		
		AbsoluteCodingSchemeVersionReferenceList csvList = new AbsoluteCodingSchemeVersionReferenceList();
		
		String csName = null;
		
		if (null != entityCodeNamespace && StringUtils.isNotEmpty(entityCodeNamespace.toString())) {
			csName = entityCodeNamespace.toString();
		}
		
		ValueDomainDefinition vdDef;
		
		try {
			// populate valueDomainDefinition ohect for supplied uri.
			vdDef = getServiceHelper().getValueDomain(valueDomainURI);
			
			if (vdDef != null && vdDef.getDefinitionEntry() != null) 
			{
				
				// resolve the value domain and populate ResolvedValueDomainCodedNodeSet, that contains the codingSchemer version information
				// plus the codedNodeSet
				ResolvedValueDomainCodedNodeSet rvdcns = getServiceHelper().getCodedNodeSetForValueDomain(vdDef, csName, csvt); 
				
				if (rvdcns != null)
				{
					CodedNodeSet cns = rvdcns.getCodedNodeSet();
					
					csvList = rvdcns.getCodingSchemeVersionRefList();
					
					ResolvedConceptReferencesIterator rcrList = cns.resolve(null, null, null, null, false);
					while (rcrList.hasNext())
					{
						ResolvedConceptReference rcr = rcrList.next();
						// now check if the supplied entityCode is part of resolved set
						if (rcr.getCode().equals(entityCode))
						{
							AbsoluteCodingSchemeVersionReference csv = Constructors.createAbsoluteCodingSchemeVersionReference(rcr.getCodingSchemeURI(), rcr.getCodingSchemeVersion());
							return csv;
						}
					}
				}				
			} else {
				md_.fatal("No Value Domain found for URI : " + valueDomainURI);
				throw new LBException("No Value Domain found for URI : " + valueDomainURI);
			}
		} catch (FindException e) {
			md_.fatal("Failed verifying concept '" + entityCode + "' in Value Domain : " + valueDomainURI , e);
			throw new LBException("Failed verifying concept '" + entityCode + "' in Value Domain : " + valueDomainURI, e);
		} catch (ServiceInitException e) {
			md_.fatal("Failed verifying concept '" + entityCode + "' in Value Domain : " + valueDomainURI, e);
			throw new LBException("Failed verifying concept '" + entityCode + "' in Value Domain : " + valueDomainURI, e);
		} catch (URISyntaxException e) {
			md_.fatal("Failed verifying concept '" + entityCode + "' in Value Domain : " + valueDomainURI, e);
			throw new LBException("Failed verifying concept '" + entityCode + "' in Value Domain : " + valueDomainURI, e);
		}
		
		// if we reach here; the supplied entityCode was not found in the valueDomain OR the codingscheme is not loaded
		return null;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#isSubDomain
	 * (java.net.URI, java.net.URI)
	 */
	public boolean isSubDomain(URI childValueDomainURI, URI parentValueDomainURI) throws LBException {
		getLogger().logMethod(new Object[] { childValueDomainURI, parentValueDomainURI });

		List<EntityReferenceWrapper> parentEntityRefs = null;
		List<String> assnRefs = null;
		CodedNodeSet childCNS = null;
		CodedNodeSet parentCNS = null;
		AbsoluteCodingSchemeVersionReferenceList parentcsvList = null;
		
		// populate valueDomainDefinitions for both child and parent value domains
		ValueDomainDefinition childVDDef = getServiceHelper().getValueDomain(childValueDomainURI);
		ValueDomainDefinition parentVDDef = getServiceHelper().getValueDomain(parentValueDomainURI);
		
		try {
			if (parentVDDef != null && parentVDDef.getDefinitionEntry() != null) {
				parentEntityRefs = getServiceHelper().getEntityReferences(parentVDDef);
	
				if (parentEntityRefs.size() == 0) {
					md_.fatal("There was a problem verifying value domain. Possible reason : "
									+ "No Coding Scheme reference found for entityCode in valueDomain : "
									+ parentValueDomainURI);
					throw new LBException("There was a problem verifying value domain. Possible reason : "
									+ "No Coding Scheme reference found for entityCode in valueDomain : "
									+ parentValueDomainURI);
				}
				
				assnRefs = getServiceHelper().getUniqueAssociationInVD(parentEntityRefs);
			}
			
			// Resolve the child value domain and populate CodedNodeSet for this domain and the coding scheme version that was used.
			ResolvedValueDomainCodedNodeSet childrvdcns = getServiceHelper().getCodedNodeSetForValueDomain(childVDDef, null, null);
			if (childrvdcns != null)
				childCNS = childrvdcns.getCodedNodeSet();
			if (childCNS == null) {
				md_.fatal("There was a problem creating CodedNodeSet for child valueDomain : " + childValueDomainURI);
				throw new LBException("There was a problem creating CodedNodeSet for child valueDomain : " + childValueDomainURI);
			}
			
			// Resolve the parent value domain and populate CodedNodeSet for this domain and the coding scheme version that was used.
			ResolvedValueDomainCodedNodeSet parentrvdcns = getServiceHelper().getCodedNodeSetForValueDomain(parentVDDef, null, null);
			if (parentrvdcns != null)
			{
				parentCNS = parentrvdcns.getCodedNodeSet();
				parentcsvList = parentrvdcns.getCodingSchemeVersionRefList();
			}
			
			if (parentCNS == null) {
				md_.fatal("There was a problem creating CodedNodeSet for parent valueDomain : " + parentValueDomainURI);
				throw new LBException("There was a problem creating CodedNodeSet for parent valueDomain : " + parentValueDomainURI);
			}
		} catch (FindException e1) {
			md_.fatal("Failed verifying isSubdomain for child Value Domain '" + childValueDomainURI + "' and patend Value Domain : " + parentValueDomainURI, e1);
			throw new LBException("Failed verifying isSubdomain for child Value Domain '" + childValueDomainURI + "' and patend Value Domain : " + parentValueDomainURI, e1);
		} catch (ServiceInitException e1) {
			md_.fatal("Failed verifying isSubdomain for child Value Domain '" + childValueDomainURI + "' and patend Value Domain : " + parentValueDomainURI, e1);
			throw new LBException("Failed verifying isSubdomain for child Value Domain '" + childValueDomainURI + "' and patend Value Domain : " + parentValueDomainURI, e1);
		} catch (URISyntaxException e1) {
			md_.fatal("Failed verifying isSubdomain for child Value Domain '" + childValueDomainURI + "' and patend Value Domain : " + parentValueDomainURI, e1);
			throw new LBException("Failed verifying isSubdomain for child Value Domain '" + childValueDomainURI + "' and patend Value Domain : " + parentValueDomainURI, e1);
		}
		
		// resolve the child codedNodeSet; don't have to resolve the entities.
		ResolvedConceptReferencesIterator childRCRItr = childCNS.resolve(null, null, null, null, false);
		
		ResolvedConceptReference childRCR = null;		
		CodedNodeGraph cng = null;
		
		// Now for each entities in child valueDomain, check if it child of parent valueDomain entity
		while(childRCRItr.hasNext())
		{
			childRCR = childRCRItr.next();			
			boolean isSubDomain = false;	
			
			if (parentCNS != null)
			{
				//for(String csName : csRefs)
				for (int i = 0; i < parentcsvList.getAbsoluteCodingSchemeVersionReferenceCount(); i++)
				{
					AbsoluteCodingSchemeVersionReference parentcsv = parentcsvList.getAbsoluteCodingSchemeVersionReference(i);
					for(String assn : assnRefs)
					{
						cng = getLexBIGService().getNodeGraph(parentcsv.getCodingSchemeURN(), Constructors.createCodingSchemeVersionOrTag(null, parentcsv.getCodingSchemeVersion()), null);
						cng.restrictToAssociations(Constructors.createNameAndValueList(assn), null);						
						cng.restrictToCodes(parentCNS);
						
						if (cng.isCodeInGraph(Constructors.createConceptReference(childRCR.getCode(), parentcsv.getCodingSchemeURN())))
						{
							isSubDomain = true;
							break;
						}
					}
				}
			}
			// if current concept from child VD was not found as a sub-class of all the concepts in parent VD, return false; 
			if (!isSubDomain)
				return false;
		}
		
		// If we reach this point; all the entities in child VD are sub-class of parent VD
		return true;
	}


	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#listValueDomains
	 * (java.lang.String)
	 */
	public URI[] listValueDomains(String valueDomainName) throws LBException {
		getLogger().logMethod(new Object[] { valueDomainName });
		URI[] uris;
		try {
			uris = getServiceHelper().getValueDomainServices().findByValueDomainName(
					valueDomainName);
		} catch (FindException e) {
			md_.fatal("Failed during list value domain for value domain name : " + valueDomainName, e);
			throw new LBException("Failed during list value domain for value domain name : " + valueDomainName, e);
		}
		
		return uris;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @seeorg.lexgrid.extension.valuedomain.LexEVSValueDomainServices#
	 * resolveValueDomain(java.net.URI,
	 * org.LexGrid.LexBIG.DataModel.Collections.
	 * AbsoluteCodingSchemeVersionReferenceList)
	 */
	public ResolvedValueDomainDefinition resolveValueDomain(URI valueDomainURI,
			AbsoluteCodingSchemeVersionReferenceList csVersionList)
			throws LBException {
		getLogger().logMethod(new Object[] { valueDomainURI, csVersionList });
		
		CodedNodeSet finalCNS = null;
		ValueDomainDefinition vdDef = null;
		
		Hashtable<String, String> csvTbl = null;
		String csName = null;
		String version = null;
		csvTbl = new Hashtable<String, String>();
		AbsoluteCodingSchemeVersionReferenceList csVersionListUsed = new AbsoluteCodingSchemeVersionReferenceList();
		
		try {
			// populate valueDomainDefinition object for supplied uri.
			vdDef = getServiceHelper().getValueDomain(valueDomainURI);			
			
			csName = vdDef.getDefaultCodingScheme();
			
			// if csVersion list if supplied; make sure it is loaded in the instance; 
			// if not loaded; use the version that is loaded. 
			// Will return the used csVersion in resolved object.
			if (csVersionList != null)
			{
				for (int i = 0; i < csVersionList.getAbsoluteCodingSchemeVersionReferenceCount(); i++)
				{
					AbsoluteCodingSchemeVersionReference acsvr = csVersionList.getAbsoluteCodingSchemeVersionReference(i);
					csName = acsvr.getCodingSchemeURN();
					version = acsvr.getCodingSchemeVersion();
					// check if this version of coding scheme is loaded, otherwise get which is loaded
					
					AbsoluteCodingSchemeVersionReferenceList acsvrLoaded = getServiceHelper().getAbsoluteCodingSchemeVersionReference(csName, Constructors.createCodingSchemeVersionOrTag(null, version));
					if (acsvrLoaded != null)
					{
						csvTbl.put(csName, version);
						csVersionListUsed.addAbsoluteCodingSchemeVersionReference(acsvr);
					}
				}
			}
			else
			{
				AbsoluteCodingSchemeVersionReferenceList acsvrLoaded = getServiceHelper().getAbsoluteCodingSchemeVersionReference(csName, null);
				for (int i = 0; i < acsvrLoaded.getAbsoluteCodingSchemeVersionReferenceCount(); i++)
				{
					AbsoluteCodingSchemeVersionReference acsvr = acsvrLoaded.getAbsoluteCodingSchemeVersionReference(i);
					csvTbl.put(acsvr.getCodingSchemeURN(), acsvr.getCodingSchemeVersion());
					csVersionListUsed.addAbsoluteCodingSchemeVersionReference(acsvr);
				}
			}
			
			// populate entityRef wrapper object for all the definitionEntry found in the VD.
			List<EntityReferenceWrapper> entityRefs = getServiceHelper().getEntityReferences(vdDef);
			
			// Resolve each entityRef object and perform the definition operator (OR, ADD, or SUBTRACT) assigned for each of them.
			for (EntityReferenceWrapper entityRef : entityRefs)
			{
				csName = entityRef.getCodingSchemeReference();
				version = null;
				if (csvTbl != null)
					version = csvTbl.get((String) csName);
				
				// populate CodedNodeSet for the DefinitionEntry
				CodedNodeSet cns = getServiceHelper().getCodedNodeSetForValueDomain(entityRef, csName, Constructors.createCodingSchemeVersionOrTag(null, version));
				
				if (cns != null)
				{
					if (finalCNS == null)
					{
						finalCNS = cns;
					}
					else
					{
						// perform the operator operation according to the operator assigned to definitionEntry
						if (entityRef.getDefinitionOperator().getLiteral().equals(DefinitionOperator.OR.name()) )
							finalCNS.union(cns);
						else if (entityRef.getDefinitionOperator().getLiteral().equals(DefinitionOperator.AND.name()) )
							finalCNS.intersect(cns);
						else if (entityRef.getDefinitionOperator().getLiteral().equals(DefinitionOperator.SUBTRACT.name()) )
							finalCNS.difference(cns);
					}
				}
			}
			
		} catch (FindException e) {
			md_.fatal("Failed during resolving value domain : " + valueDomainURI, e);
			throw new LBException("Failed during resolving value domain : " + valueDomainURI, e);
		} catch (ServiceInitException e) {
			md_.fatal("Failed during resolving value domain : " + valueDomainURI, e);
			throw new LBException("Failed during resolving value domain : " + valueDomainURI, e);
		} catch (URISyntaxException e) {
			md_.fatal("Failed during resolving value domain : " + valueDomainURI, e);
			throw new LBException("Failed during resolving value domain : " + valueDomainURI, e);
		}
		
		ResolvedValueDomainDefinition rvddef = null;
		// Now populate ResolvedValueDomainDefinition object
		if (finalCNS != null)
		{
			rvddef = new ResolvedValueDomainDefinition();
			rvddef.setDefaultCodingScheme(vdDef.getDefaultCodingScheme());
			rvddef.setRepresentsRealmOrContext(vdDef.getRepresentsRealmOrContext());
			rvddef.setSource(vdDef.getSource());
			rvddef.setValueDomainName(vdDef.getValueDomainName());
			rvddef.setValueDomainURI(valueDomainURI);
			rvddef.setCodingSchemeVersionRefList(csVersionListUsed);
			rvddef.setResolvedConceptReferenceIterator(finalCNS.resolve(null, null, null));
		}
		
		return rvddef;
	}

	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#isDomain(java.lang.String, java.lang.String, org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag)
	 */
	public boolean isDomain(String domainId, String codingSchemeName,
			CodingSchemeVersionOrTag csvt) throws LBException {
		getLogger().logMethod(new Object[] { domainId, codingSchemeName, csvt });
		
		LocalNameList lnl = Constructors.createLocalNameList(SQLTableConstants.ENTRY_STATE_TYPE_VALUEDOMAIN);
		
		// Populate CodedNodeSet for supplied domainId and csVersion and restrict to 'valueDomain' type.
		CodedNodeSet cns = getLexBIGService().getNodeSet(codingSchemeName, csvt, lnl);
		cns.restrictToCodes(Constructors.createConceptReferenceList(domainId));
		
		Iterator<ResolvedConceptReference> rcrIter = cns.resolveToList(null, null, null, null, 1).iterateResolvedConceptReference();
		// If there were entities after resolving the codedNodeSet; the supplied domainId is a valueDomain
		if (rcrIter.hasNext()) {
			return true;
		}
		return false;
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#removeValueDomain(java.net.URI)
	 */
	public void removeValueDomain(URI valueDomainURI) throws LBException, RemoveException {
		try {
			getServiceHelper().getValueDomainServices().remove(valueDomainURI);
		} catch (FindException e) {
			md_.fatal("Failed during removing value domain : " + valueDomainURI, e);
			throw new LBException("Failed during removing value domain : " + valueDomainURI, e);
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#removeAllValueDomains()
	 */
	public void removeAllValueDomains() throws RemoveException, LBException{
		try {
			URI[] vdURIs = listValueDomains(null);
			
			for (URI vdURI : vdURIs)
			{
				getServiceHelper().getValueDomainServices().remove(vdURI);
			}
		} catch (FindException e) {
			md_.fatal("Failed during removing all value domains", e);
			throw new LBException("Failed during removing all value domains", e);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.lexgrid.extension.valuedomain.LexEVSValueDomainServices#dropValueDomainTables()
	 */
	public void dropValueDomainTables() throws RemoveException, LBException {
		try {
			getServiceHelper().getValueDomainServices().dropValueDomainTables(false);
		} catch (FindException e) {
			md_.fatal("Failed to drop value domain tables", e);
			throw new LBException("Failed to drop value domain tables", e);
		} catch (SQLException e) {
			md_.fatal("Failed to drop value domain tables", e);
			throw new LBException("Failed to drop value domain tables", e);
		}
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see org.LexGrid.LexBIG.Extensions.Extendable#getDescription()
	 */
	public String getDescription() {
		return desc_;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.LexGrid.LexBIG.Extensions.Extendable#getName()
	 */
	public String getName() {
		return name_;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.LexGrid.LexBIG.Extensions.Extendable#getProvider()
	 */
	public String getProvider() {
		return provider_;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.LexGrid.LexBIG.Extensions.Extendable#getVersion()
	 */
	public String getVersion() {
		return version_;
	}

	/**
	 * Assign the associated LexBIGService instance.
	 * <p>
	 * Note: This method must be invoked by users of the distributed LexBIG API
	 * to set the service to an EVSApplicationService object, allowing client
	 * side implementations to use these convenience methods.
	 */
	@LgClientSideSafe
	public void setLexBIGService(LexBIGService lbs) {
		lbs_ = lbs;
	}

	/**
	 * Return the associated LexBIGService instance; lazy initialized as
	 * required.
	 */
	@LgClientSideSafe
	public LexBIGService getLexBIGService() {
		if (lbs_ == null)
			lbs_ = LexBIGServiceImpl.defaultInstance();
		return lbs_;
	}

	private VDServices getVDService() throws ServiceInitException, LBException {
		if (vds_ == null)
			vds_ = getServiceHelper().getValueDomainServices();
		return vds_;
	}

	private VDSServices getVDSService() throws ServiceInitException {
		if (vdss_ == null)
			vdss_ = getServiceHelper().getValueDomainsServices();
		return vdss_;
	}
	
	private VDServiceHelper getServiceHelper(){
		if (sh_ == null)
		{
			SystemVariables sv = ResourceManager.instance().getSystemVariables();
			try {
				sh_ = new VDServiceHelper(sv.getAutoLoadDBURL(), sv.getAutoLoadDBDriver(), sv.getAutoLoadDBUsername(),
						sv.getAutoLoadDBPassword(), sv.getAutoLoadDBPrefix(), true, md_);
			} catch (LBParameterException e) {
				md_.fatal("Problem getting ServiceHelper", e);
				e.printStackTrace();
			} catch (LBInvocationException e) {
				md_.fatal("Problem getting ServiceHelper", e);
				e.printStackTrace();
			}
		}
		return sh_;
	}
	
	private String getStringFromURI(URI uri) throws LBParameterException {
		if ("file".equals(uri.getScheme()))

        {
            File temp = new File(uri);
            return temp.getAbsolutePath();
        }
		return uri.toString();
    }
}