/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.valuedomain.persistence;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import org.LexGrid.LexBIG.DataModel.Collections.AbsoluteCodingSchemeVersionReferenceList;
import org.LexGrid.LexBIG.DataModel.Collections.AssociatedConceptList;
import org.LexGrid.LexBIG.DataModel.Collections.CodingSchemeRenderingList;
import org.LexGrid.LexBIG.DataModel.Collections.ConceptReferenceList;
import org.LexGrid.LexBIG.DataModel.Collections.NameAndValueList;
import org.LexGrid.LexBIG.DataModel.Collections.ResolvedConceptReferenceList;
import org.LexGrid.LexBIG.DataModel.Core.AbsoluteCodingSchemeVersionReference;
import org.LexGrid.LexBIG.DataModel.Core.AssociatedConcept;
import org.LexGrid.LexBIG.DataModel.Core.Association;
import org.LexGrid.LexBIG.DataModel.Core.CodingSchemeVersionOrTag;
import org.LexGrid.LexBIG.DataModel.Core.ConceptReference;
import org.LexGrid.LexBIG.DataModel.InterfaceElements.CodingSchemeRendering;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBInvocationException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.Exceptions.LBResourceUnavailableException;
import org.LexGrid.LexBIG.Impl.LexBIGServiceImpl;
import org.LexGrid.LexBIG.Impl.dataAccess.ResourceManager;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.LexBIG.Utility.Constructors;
import org.LexGrid.LexBIG.Utility.ConvenienceMethods;
import org.LexGrid.LexBIG.Utility.Iterators.ResolvedConceptReferencesIterator;
import org.LexGrid.codingSchemes.CodingScheme;
import org.LexGrid.emf.valueDomains.CodingSchemeReference;
import org.LexGrid.emf.valueDomains.DefinitionEntry;
import org.LexGrid.emf.valueDomains.EntityReference;
import org.LexGrid.emf.valueDomains.ValueDomainDefinition;
import org.LexGrid.emf.valueDomains.persistence.ValueDomainHome;
import org.LexGrid.emf.valueDomains.persistence.ValueDomainsHome;
import org.LexGrid.managedobj.FindException;
import org.LexGrid.managedobj.HomeServiceBroker;
import org.LexGrid.managedobj.ServiceInitException;
import org.LexGrid.managedobj.ServiceUnavailableException;
import org.LexGrid.managedobj.jdbc.JDBCConnectionDescriptor;
import org.LexGrid.messaging.LgMessageDirectorIF;
import org.LexGrid.naming.Mappings;
import org.LexGrid.naming.SupportedAssociation;
import org.LexGrid.valueDomains.types.DefinitionOperator;
import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;
import org.lexgrid.valuedomain.dto.EntityReferenceWrapper;
import org.lexgrid.valuedomain.dto.ResolvedValueDomainCodedNodeSet;

import edu.mayo.informatics.lexgrid.convert.utility.Constants;
import edu.mayo.informatics.lexgrid.convert.utility.MessageRedirector;

/**
 * Helper class for Value Domain extension.
 * 
 * @author <A HREF="mailto:dwarkanath.sridhar@mayo.edu">Sridhar Dwarkanath</A>
 */
public class VDServiceHelper {
	private static Logger log = Logger.getLogger("convert.SQL");

	private String sqlServer, sqlDriver, sqlUsername, sqlPassword, tablePrefix;
	private boolean failOnAllErrors;
	private HomeServiceBroker broker_ = null;
	private ValueDomainHome vdContext_ = null;
	private ValueDomainsHome vdsContext_ = null;
	private LgMessageDirectorIF md_;
	private LexBIGService lbs_;
	private VDServices vds_;
	private VDSServices vdss_;
	private PickListServices pls_;
	private PickListsServices plss_;
	private ConvenienceMethods cm_;

	private Hashtable<String, String> csvtAssnsToAssnCode_;

	public VDServiceHelper(String sqlServer, String sqlDriver,
			String sqlUsername, String sqlPassword, String tablePrefix,
			boolean failOnAllErrors, LgMessageDirectorIF messages)
			throws LBParameterException, LBInvocationException {
		this.sqlServer = sqlServer;
		this.sqlDriver = sqlDriver;
		this.sqlUsername = sqlUsername;
		this.sqlPassword = sqlPassword;
		this.tablePrefix = tablePrefix;
		this.md_ = messages;
		this.failOnAllErrors = failOnAllErrors;
		cm_ = new ConvenienceMethods();
	}

	/**
	 * Returns a SQL-based services to load, manage, and persist valueDomain
	 * objects.
	 * 
	 * @return ValueDomainService
	 * @throws ServiceInitException
	 */
	private VDSServices getVDSService() throws ServiceInitException {
		if (vdss_ != null)
			return vdss_;
		
		HomeServiceBroker broker = getBroker();
		if (vdsContext_ != null)
			try {
				vdss_ = (VDSServices) broker.getService(vdsContext_);
			} catch (ServiceUnavailableException sue) {
			}
		if (vdss_ == null) {
			JDBCConnectionDescriptor desc = new JDBCConnectionDescriptor();
			desc.setDbUid(sqlUsername);
			desc.setDbPwd(sqlPassword);
			desc.setDbUrl(sqlServer);
			desc.setUseUTF8(true);
			desc.setAutoRetryFailedConnections(true);
			try {
				desc.setDbDriver(sqlDriver);
			} catch (ClassNotFoundException e) {
				md_.fatal("The requested data base driver was not found on the classpath");
				throw new ServiceInitException("The requested driver was not found on the classpath");
			}
			vdss_ = new VDSServices(broker, desc, false, tablePrefix,
					failOnAllErrors, new MessageRedirector(md_), false);
			vdss_.setPageSize(Constants.mySqlBatchSize);
			vdsContext_ = new ValueDomainsHome(broker);
			broker.registerService(vdsContext_, vdss_);
		}
		return vdss_;
	}
	
	/**
	 * Returns a SQL-based services to load, manage, and persist pick list
	 * objects.
	 * 
	 * @return PickListsService
	 * @throws ServiceInitException
	 */
	public PickListsServices getPickListsService() throws ServiceInitException {
		if (plss_ != null)
			return plss_;
		
		HomeServiceBroker broker = getBroker();
		if (vdsContext_ != null)
			try {
				plss_ = (PickListsServices) broker.getService(vdsContext_);
			} catch (ServiceUnavailableException sue) {
			}
		if (plss_ == null) {
			JDBCConnectionDescriptor desc = new JDBCConnectionDescriptor();
			desc.setDbUid(sqlUsername);
			desc.setDbPwd(sqlPassword);
			desc.setDbUrl(sqlServer);
			desc.setUseUTF8(true);
			desc.setAutoRetryFailedConnections(true);
			try {
				desc.setDbDriver(sqlDriver);
			} catch (ClassNotFoundException e) {
				md_.fatal("The requested data base driver was not found on the classpath");
				throw new ServiceInitException("The requested driver was not found on the classpath");
			}
			plss_ = new PickListsServices(broker, desc, false, tablePrefix,
					failOnAllErrors, new MessageRedirector(md_), false);
			plss_.setPageSize(Constants.mySqlBatchSize);
			vdsContext_ = new ValueDomainsHome(broker);
			broker.registerService(vdsContext_, plss_);
		}
		return plss_;
	}
	
	/**
	 * Returns a SQL-based services to load, manage, and persist pick list
	 * objects.
	 * 
	 * @return PickListServices
	 * @throws ServiceInitException
	 */
	public PickListServices getPickListService() throws ServiceInitException {
		if (pls_ != null)
			return pls_;
		
		HomeServiceBroker broker = getBroker();
		if (vdContext_ != null)
			try {
				pls_ = (PickListServices) broker.getService(vdContext_);
			} catch (ServiceUnavailableException sue) {
			}
		if (pls_ == null) {
			JDBCConnectionDescriptor desc = new JDBCConnectionDescriptor();
			desc.setDbUid(sqlUsername);
			desc.setDbPwd(sqlPassword);
			desc.setDbUrl(sqlServer);
			desc.setUseUTF8(true);
			desc.setAutoRetryFailedConnections(true);
			try {
				desc.setDbDriver(sqlDriver);
			} catch (ClassNotFoundException e) {
				md_.fatal("The requested data base driver was not found on the classpath");
				throw new ServiceInitException("The requested driver was not found on the classpath");
			}
			pls_ = new PickListServices(broker, desc, false, tablePrefix,
					failOnAllErrors, new MessageRedirector(md_), false);
			pls_.setPageSize(Constants.mySqlBatchSize);
			vdContext_ = new ValueDomainHome(broker);
			broker.registerService(vdContext_, pls_);
		}
		return pls_;
	}

	/**
	 * Returns a SQL-based services to load, manage, and persist valueDomain
	 * objects.
	 * 
	 * @return ValueDomainService
	 * @throws ServiceInitException
	 */
	private VDServices getVDDefinitionService() throws ServiceInitException {
		if (vds_ != null)
			return vds_;
		
		HomeServiceBroker broker = getBroker();
		vds_ = null;
		if (vdContext_ != null)
			try {
				vds_ = (VDServices) broker.getService(vdContext_);
			} catch (ServiceUnavailableException sue) {
			}
		if (vds_ == null) {
			JDBCConnectionDescriptor desc = new JDBCConnectionDescriptor();
			desc.setDbUid(sqlUsername);
			desc.setDbPwd(sqlPassword);
			desc.setDbUrl(sqlServer);
			desc.setUseUTF8(true);
			desc.setAutoRetryFailedConnections(true);
			try {
				desc.setDbDriver(sqlDriver);
			} catch (ClassNotFoundException e) {
				md_.fatal("The requested data base driver was not found on the classpath");
				throw new ServiceInitException("The requested driver was not found on the classpath");
			}
			vds_ = new VDServices(broker, desc, false, tablePrefix,
					failOnAllErrors, new MessageRedirector(md_), false);
			vds_.setPageSize(Constants.mySqlBatchSize);
			vdContext_ = new ValueDomainHome(broker);
			broker.registerService(vdContext_, vds_);
		}
		return vds_;
	}

	/**
	 * Returns the broker used to manage SQL services.
	 * 
	 * @return HomeServiceBroker
	 */
	protected HomeServiceBroker getBroker() {
		if (broker_ == null)
			broker_ = new HomeServiceBroker();
		return broker_;
	}
	
	/**
	 * Returns the valueDomainDefinition object for supplied value domain URI. 
	 * @param valueDomainURI
	 * @return ValueDomainDefinition if found; otherwise NULL
	 * @throws LBException
	 */
	public ValueDomainDefinition getValueDomain(URI valueDomainURI) throws LBException {
		ValueDomainDefinition vdDef = null;
		try {
			vdDef = (ValueDomainDefinition) getValueDomainServices().findByPrimaryKey(valueDomainURI);
		} catch (FindException e) {
			md_.fatal("Failed while processing value domain definition : " + valueDomainURI, e);
			throw new LBException(e.getMessage());
		}
		return vdDef;
	}

	/**
	 * Returns all coding schemes referenced in the supplied valueDomainDefinition
	 * @param vdDef
	 * @return List of CodingSchemeReference
	 * @throws FindException
	 * @throws ServiceInitException
	 * @throws LBException
	 */
	public List<CodingSchemeReference> getCodingSchemeReferences(ValueDomainDefinition vdDef) 
		throws FindException, ServiceInitException, LBException {
		List<CodingSchemeReference> csRefs = null;
		
		if (vdDef != null && vdDef.getDefinitionEntry() != null) {
			Iterator<DefinitionEntry> iter = vdDef.getDefinitionEntry().iterator();
			while (iter.hasNext()) {
				if (csRefs == null)
					csRefs = new ArrayList<CodingSchemeReference>();
				
				DefinitionEntry de = iter.next();
				csRefs.add(de.getCodingSchemeReference());
			}
		}
		return csRefs;
	}

	/**
	 * Returns name of the codingScheme in the supplied value domain URI and for the entityCode
	 * @param valueDomainURI
	 * @param entityCode
	 * @return String CodingScheme name
	 * @throws FindException
	 * @throws ServiceInitException
	 * @throws LBException
	 */
	public String getCodingSchemeName(URI valueDomainURI, String entityCode)
			throws FindException, ServiceInitException, LBException {

		String csName = null;

		ValueDomainDefinition vdDef = getValueDomain(valueDomainURI);

		if (vdDef != null && vdDef.getDefaultCodingScheme() != null) {
			csName = vdDef.getDefaultCodingScheme();
			if (StringUtils.isNotEmpty(csName))
				return csName;
		}

		if (vdDef != null && vdDef.getDefinitionEntry() != null) {
			Iterator<DefinitionEntry> iter = vdDef.getDefinitionEntry().iterator();
			while (iter.hasNext()) {
				DefinitionEntry de = iter.next();
				if (StringUtils.isNotEmpty(entityCode)) {
					if (de.getEntityReference().getEntityCode().equals(entityCode)) {
						csName = de.getCodingSchemeReference() == null ? null
								: de.getCodingSchemeReference().getCodingScheme();
						if (StringUtils.isNotEmpty(csName))
							return csName;
					}
				} else {
					csName = de.getCodingSchemeReference() == null ? null : de.getCodingSchemeReference().getCodingScheme();
					if (StringUtils.isNotEmpty(csName))
						return csName;
				}
			}
		}
		return csName;
	}

	/**
	 * Checks if the supplied codingScheme name/uri and the version is loaded.
	 * If the supplied version of codingScheme is loaded; checks is other versions are
	 * loaded and returns it.
	 * @param codingSchemeName
	 * @param csvt
	 * @return AbsoluteCodingSchemeVersionReferenceList list of codingScheme and version
	 * @throws LBException
	 */
	public AbsoluteCodingSchemeVersionReferenceList getAbsoluteCodingSchemeVersionReference(
			String codingSchemeName, CodingSchemeVersionOrTag csvt) throws LBException {
		AbsoluteCodingSchemeVersionReferenceList acsvrList = null;
		AbsoluteCodingSchemeVersionReference acsvr = null;
		List<String> versionsList = new ArrayList<String>();
		
		String version = null;
		String csName = codingSchemeName;
		ResourceManager rm = ResourceManager.instance();			
		
		if (csvt != null)
		{
			// check if supplied version for the coding scheme is loaded
			csName = rm.getExternalCodingSchemeNameForUserCodingSchemeNameOrId(codingSchemeName, csvt.getVersion());
			if (StringUtils.isNotEmpty(csName)){
				version = csvt.getVersion();
				acsvr = new AbsoluteCodingSchemeVersionReference();
				acsvr.setCodingSchemeURN(codingSchemeName);
				acsvr.setCodingSchemeVersion(version);
				if (acsvrList == null)
					acsvrList = new AbsoluteCodingSchemeVersionReferenceList();
				acsvrList.addAbsoluteCodingSchemeVersionReference(acsvr);
				return acsvrList;
			}
		}
		
		if (StringUtils.isEmpty(version))
		{
			CodingSchemeRenderingList csrList = getLexBIGService().getSupportedCodingSchemes();
			
			for (int i = 0; i <csrList.getCodingSchemeRenderingCount(); i++)
			{
				CodingSchemeRendering csr = csrList.getCodingSchemeRendering(i);
				if (csr.getCodingSchemeSummary().getFormalName() != null && csr.getCodingSchemeSummary().getFormalName().equals(csName))
				{
					version = csr.getCodingSchemeSummary().getRepresentsVersion();
				}
				else if (csr.getCodingSchemeSummary().getLocalName() != null && csr.getCodingSchemeSummary().getLocalName().equals(csName))
				{
					version = csr.getCodingSchemeSummary().getRepresentsVersion();
				}
				else if (csr.getCodingSchemeSummary().getCodingSchemeURI().equals(csName))
				{
					version = csr.getCodingSchemeSummary().getRepresentsVersion();
				}
				if (StringUtils.isNotBlank(version) && !versionsList.contains(version))
				{
					acsvr = new AbsoluteCodingSchemeVersionReference();
					acsvr.setCodingSchemeURN(codingSchemeName);
					acsvr.setCodingSchemeVersion(version);
					if (acsvrList == null)
						acsvrList = new AbsoluteCodingSchemeVersionReferenceList();
					acsvrList.addAbsoluteCodingSchemeVersionReference(acsvr);
					versionsList.add(version);
				}
			}
		}
		
		return acsvrList;
	}
	
	/**
	 * Checks if the supplied codingScheme and version is loaded.
	 * @param codingSchemeName
	 * @param version
	 * @return True; if given codingScheme/version is loader, otherwise, False.
	 * @throws LBException
	 */
	public boolean isCodingSchemeVersionLoaded(
			String codingSchemeName, String version) throws LBException {
		
		String csName = codingSchemeName;
		ResourceManager rm = ResourceManager.instance();			
		
		if (version != null)
		{
			// check if supplied version for the coding scheme is loaded
			csName = rm.getExternalCodingSchemeNameForUserCodingSchemeNameOrId(codingSchemeName, version);
			if (StringUtils.isNotEmpty(csName)){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Returns entityCode associated with the associationName.
	 * 
	 * @param codingScheme
	 * @param csvt
	 * @param associationName
	 * @return String entityCode for given associationName.
	 * @throws LBException
	 */
	public String getAssociationEntityCode(String codingScheme, CodingSchemeVersionOrTag csvt, String associationName)
			throws LBException {
		String associationCode = associationName;

		if (csvtAssnsToAssnCode_ == null)
			csvtAssnsToAssnCode_ = new Hashtable<String, String>();

		String key = codingScheme + (csvt != null ? csvt.getVersion() : "") + associationName;
		if (csvtAssnsToAssnCode_.containsKey(key)) {
			return csvtAssnsToAssnCode_.get(key);
		} else {
			SupportedAssociation[] suppAssns = getSupportedAssociations(codingScheme, csvt);

			// Check if the supplied association is associationName
			for (SupportedAssociation suppAssn : suppAssns) {
				if (suppAssn.getContent().equalsIgnoreCase(associationName)) {
					String value = suppAssn.getLocalId();
					csvtAssnsToAssnCode_.put(key, value);
					return value;
				}
			}
		}

		return associationCode;
	}

	/**
	 * Returns all supportedAssociation for the supplied codingScheme and version.
	 * 
	 * @param codingScheme
	 * @param versionOrTag
	 * @return Array of SupportedAssociation
	 * @throws LBException
	 */
	protected SupportedAssociation[] getSupportedAssociations(String codingScheme, CodingSchemeVersionOrTag versionOrTag)
			throws LBException {

		CodingScheme cs = getLexBIGService().resolveCodingScheme(codingScheme, versionOrTag);
		if (cs == null) {
			throw new LBResourceUnavailableException("Coding scheme not found.");
		}
		Mappings mappings = cs.getMappings();
		return mappings.getSupportedAssociation();
	}

	/**
	 * Checks if supplied entity code is present in codingScheme and version.
	 * 
	 * @param entityCode
	 * @param codingSchemeName
	 * @param csvt
	 * @return TRUE if found; FALSE otherwise
	 * @throws LBException
	 */
	public AbsoluteCodingSchemeVersionReferenceList isEntityInCodingScheme(String entityCode, String codingSchemeName,
			CodingSchemeVersionOrTag csvt) throws LBException {
		AbsoluteCodingSchemeVersionReferenceList acsvrList = null;
		CodedNodeSet cns = cm_.createCodedNodeSet(new String[] { entityCode }, codingSchemeName, csvt);
		ResolvedConceptReferencesIterator rcri = cns.resolve(null, null, null);

		if (rcri.hasNext()) {
			acsvrList = getAbsoluteCodingSchemeVersionReference(codingSchemeName, csvt);
		}
		return acsvrList;
	}
	
	
	public ResolvedConceptReferenceList getNodes(String codingScheme, CodingSchemeVersionOrTag versionOrTag, ConceptReference cr,
			String relationContext, String association, NameAndValueList associationQualifiers) throws LBException {

		CodedNodeGraph cng = getLexBIGService().getNodeGraph(codingScheme, versionOrTag, relationContext);
		if ((association != null && association.length() > 0)
				|| (associationQualifiers != null)) {
			cng = cng.restrictToAssociations(ConvenienceMethods.createNameAndValueList(association), associationQualifiers);
		}

		return cng
				.resolveAsList(cr, true, false, -1, -1, null, null, null, 500);
	}

	public Association getAssociationForwardOneLevel(String conceptCode, String relationContainerName, String association, String codingScheme,
            CodingSchemeVersionOrTag versionOrTag, boolean buildReferencedEntries, NameAndValueList associationQualifiers) throws LBException {
		return cm_.getAssociationForwardOneLevel(conceptCode, relationContainerName, association, codingScheme, versionOrTag, buildReferencedEntries, associationQualifiers);
	}
	
	public Association getAssociationReverseOneLevel(String conceptCode, String relationContainerName, String association,
			String codingScheme, CodingSchemeVersionOrTag versionOrTag, boolean buildReferencedEntries,
			NameAndValueList associationQualifiers) throws LBException {
		return cm_.getAssociationReverseOneLevel(conceptCode, relationContainerName, association, codingScheme, versionOrTag,
				buildReferencedEntries, associationQualifiers);
	}
	
	/**
	 * Returns list of all the leafNodes for the supplied entityCode in codingScheme , version and using supplied association.
	 * The leafNodes will be determined based on the direction of an assication (sourceToTarget OR targetToSource)
	 * 
	 * @param conceptCode
	 * @param relationContainerName
	 * @param association
	 * @param codingScheme
	 * @param versionOrTag
	 * @param buildReferencedEntries
	 * @param associationQualifiers
	 * @param targetToSource
	 * @return List of leaf EntityCodes.
	 * @throws LBException
	 */
	public List<String> getLeafNodes(String conceptCode, String relationContainerName, String association, String codingScheme,
            CodingSchemeVersionOrTag versionOrTag, boolean buildReferencedEntries,
            NameAndValueList associationQualifiers, boolean targetToSource) throws LBException{
		return getLeafNodes(conceptCode, relationContainerName, association, codingScheme, versionOrTag, buildReferencedEntries, associationQualifiers, new ArrayList<String>(), targetToSource);
	}
            
	/**
	 * Returns list of all the leafNodes for the supplied entityCode in codingScheme , version and using supplied association.
	 * The leafNodes will be determined based on the direction of an assication (sourceToTarget OR targetToSource)
	 * 
	 * @param conceptCode
	 * @param relationContainerName
	 * @param association
	 * @param codingScheme
	 * @param versionOrTag
	 * @param buildReferencedEntries
	 * @param associationQualifiers
	 * @param leafNodes
	 * @param targetToSource
	 * @return
	 * @throws LBException
	 */	
	private List<String> getLeafNodes(String conceptCode, String relationContainerName, String association, String codingScheme,
            CodingSchemeVersionOrTag versionOrTag, boolean buildReferencedEntries,
            NameAndValueList associationQualifiers, ArrayList<String> leafNodes, boolean targetToSource) throws LBException{
		Association assn = null;
			
		if (targetToSource)
		{
			assn = getAssociationReverseOneLevel(conceptCode, relationContainerName, association, codingScheme,
		            versionOrTag, buildReferencedEntries, associationQualifiers);
			
			if (assn != null)
			{
				AssociatedConceptList acList = assn.getAssociatedConcepts();
				for (int i = 0; i < acList.getAssociatedConceptCount(); i++)
				{
					AssociatedConcept ac = acList.getAssociatedConcept(i);
					String cc = ac.getCode();
					if (StringUtils.isNotEmpty(cc))
						getLeafNodes(cc, relationContainerName, association, codingScheme,
					            versionOrTag, buildReferencedEntries, associationQualifiers, leafNodes, targetToSource);
				}
			}
			else
			{
				leafNodes.add(conceptCode);
			}
		}
		else
		{
			assn = getAssociationForwardOneLevel(conceptCode, relationContainerName, association, codingScheme,
		            versionOrTag, buildReferencedEntries, associationQualifiers);
			
			if (assn != null)
			{
				AssociatedConceptList acList = assn.getAssociatedConcepts();
				for (int i = 0; i < acList.getAssociatedConceptCount(); i++)
				{
					AssociatedConcept ac = acList.getAssociatedConcept(i);
					String cc = ac.getCode();
					if (StringUtils.isNotEmpty(cc))
						getLeafNodes(cc, relationContainerName, association, codingScheme,
					            versionOrTag, buildReferencedEntries, associationQualifiers, leafNodes, targetToSource);
				}
			}
			else
			{
				leafNodes.add(conceptCode);
			}
		}		
		
		return leafNodes;
	}
	
	/**
	 * Resolves the supplied valueDomainDefinition object and returns the CodedNodeSet plus the codingScheme and version used.
	 * @param vdDef
	 * @param csName
	 * @param csvt
	 * @return ResolvedValueDomainCodedNodeSet
	 * @throws LBException
	 * @throws FindException
	 * @throws ServiceInitException
	 * @throws URISyntaxException
	 */
	public ResolvedValueDomainCodedNodeSet getCodedNodeSetForValueDomain(ValueDomainDefinition vdDef, String csName, CodingSchemeVersionOrTag csvt) throws LBException, FindException, ServiceInitException, URISyntaxException{
		CodedNodeSet finalCNS = null;		
		AbsoluteCodingSchemeVersionReferenceList usedCSVList = new AbsoluteCodingSchemeVersionReferenceList();
		List<String> csNames = new ArrayList<String>();
		ResolvedValueDomainCodedNodeSet rvdcns = null;
		
		if (vdDef != null)
		{
			if (vdDef.getDefinitionEntry() != null) 
			{
				// Populate entityReference wrapper object for each definitionEntries found in thie value domain
				List<EntityReferenceWrapper> entityRefs = getEntityReferences(vdDef);
	
				if (entityRefs.size() == 0) {
					throw new LBException("There was a problem processing entries in valueDomain : "
									+ vdDef.getValueDomainURI());
				}
				
				// For each entityReference; populae the codedNodeSet and apply the operator assigned with the definitionEntry
				for (int i = 0; i < entityRefs.size(); i++) {
					EntityReferenceWrapper entityRef = entityRefs.get(i);
					
					String vdCSRef = entityRef.getCodingSchemeReference();
					CodingSchemeVersionOrTag vdCSV = null;
					if (vdCSRef.equalsIgnoreCase(csName))
					{
						if (csvt != null)
						{
							if (isCodingSchemeVersionLoaded(csName, csvt.getVersion()))
								vdCSV = csvt;
						}
					}
					
					if (vdCSV == null)
					{
						AbsoluteCodingSchemeVersionReferenceList csList = getAbsoluteCodingSchemeVersionReference(vdCSRef, null);
						// if we find csv loaded, take the first one found
						if (csList != null && csList.getAbsoluteCodingSchemeVersionReferenceCount() > 0)	
						{
							vdCSV = Constructors.createCodingSchemeVersionOrTag(null, csList.getAbsoluteCodingSchemeVersionReference(0).getCodingSchemeVersion());
						}							
					}
					
					CodedNodeSet cns = getCodedNodeSetForValueDomain(entityRef, vdCSRef, vdCSV);
					
					if (cns != null)
					{
						if (finalCNS == null)
						{
							finalCNS = cns;
						}
						else
						{
							// Perform the operator operation based on the operator associated with this definitionEntry
							if (entityRef.getDefinitionOperator().getLiteral().equals(DefinitionOperator.OR.name()) )
								finalCNS.union(cns);
							else if (entityRef.getDefinitionOperator().getLiteral().equals(DefinitionOperator.AND.name()) )
								finalCNS.intersect(cns);
							else if (entityRef.getDefinitionOperator().getLiteral().equals(DefinitionOperator.SUBTRACT.name()) )
								finalCNS.difference(cns);
						}
						
						// Add the coding scheme name and version used in the codedNodeSet
						if (!csNames.contains(csName))
						{
							AbsoluteCodingSchemeVersionReference csv = Constructors.createAbsoluteCodingSchemeVersionReference(vdCSRef, vdCSV.getVersion());
							usedCSVList.addAbsoluteCodingSchemeVersionReference(csv);
							csNames.add(csName);
						}
					}
				}
			}
			else
			{
				// if there were no definitionEntry; the valueDomain is referencing entire codingScheme
				if (StringUtils.isEmpty(csName))
					csName = vdDef.getDefaultCodingScheme();
				
				AbsoluteCodingSchemeVersionReferenceList csList = getAbsoluteCodingSchemeVersionReference(csName, null);
				// if we find csv loaded, take the first one found
				if (csList.getAbsoluteCodingSchemeVersionReferenceCount() > 0)
				{
					AbsoluteCodingSchemeVersionReference csv = Constructors.createAbsoluteCodingSchemeVersionReference(csName, csList.getAbsoluteCodingSchemeVersionReference(0).getCodingSchemeVersion());
					usedCSVList.addAbsoluteCodingSchemeVersionReference(csv);
					finalCNS = getLexBIGService().getCodingSchemeConcepts(csName, Constructors.createCodingSchemeVersionOrTag(null, csList.getAbsoluteCodingSchemeVersionReference(0).getCodingSchemeVersion()));
				}
			}
		}
		
		if (finalCNS != null)
		{
			rvdcns = new ResolvedValueDomainCodedNodeSet();
			rvdcns.setCodedNodeSet(finalCNS);
			rvdcns.setCodingSchemeVersionRefList(usedCSVList);
		}
		
		return rvdcns;
	}
	
	/**
	 * Resolves the supplied valueDomainDefinition object and returns the CodedNodeSet plus the codingScheme and version used.
	 * @param entityRef
	 * @param csName
	 * @param csvt
	 * @return CodedNodeSet
	 * @throws LBException
	 * @throws FindException
	 * @throws ServiceInitException
	 * @throws URISyntaxException
	 */
	public CodedNodeSet getCodedNodeSetForValueDomain(EntityReferenceWrapper entityRef, String csName, CodingSchemeVersionOrTag csvt) throws LBException, FindException, ServiceInitException, URISyntaxException{
		CodedNodeSet cns = null;
				
		if (StringUtils.isEmpty(csName))
			csName = entityRef.getCodingSchemeReference();
				
		if (entityRef.isLeafOnly())
		{
			List<String> nodes = getLeafNodes(entityRef.getEntityCode(), null, entityRef.getReferenceAssociation(), csName, csvt, false, null, entityRef.isTargetToSource());
			ConceptReferenceList cList = new ConceptReferenceList();
			for (String node : nodes)
			{
				cList.addConceptReference(Constructors.createConceptReference(node, csName));
			}
			if (cList.getConceptReferenceCount() > 0)
				cns = getLexBIGService().getCodingSchemeConcepts(csName, csvt);
					
			cns.restrictToCodes(cList);
		}
		else
		{
			cns = cm_.createCodedNodeSet(new String[] {entityRef.getEntityCode()}, csName, csvt);
			
			if (!entityRef.isTransitiveClosure())
			{
				Association assn = null;
						
				if (entityRef.isTargetToSource())
					assn = getAssociationReverseOneLevel(entityRef.getEntityCode(), null, entityRef.getReferenceAssociation(), csName, csvt, false, null);
				else
					assn = getAssociationForwardOneLevel(entityRef.getEntityCode(), null, entityRef.getReferenceAssociation(), csName, csvt, false, null);
						
				if (assn != null)
				{
					Iterator<AssociatedConcept> acList = assn.getAssociatedConcepts().iterateAssociatedConcept();
					while (acList.hasNext()){
						AssociatedConcept ac = acList.next();
						cns.union(cm_.createCodedNodeSet(new String[] {ac.getCode()}, csName, csvt));
					}
				}
			}
			else
			{
				CodedNodeGraph cng = getLexBIGService().getNodeGraph(csName, csvt, null);
				cng.restrictToAssociations(Constructors.createNameAndValueList(entityRef.getReferenceAssociation()), null);
				
				if (entityRef.isTargetToSource())
					cns.union(cng.toNodeList(Constructors.createConceptReference(entityRef.getEntityCode(), csName), false, true, -1, -1));
				else
					cns.union(cng.toNodeList(Constructors.createConceptReference(entityRef.getEntityCode(), csName), true, false, -1, -1));
			}
		}
				
		return cns;
	}
	
	/**
	 * Returns all unique associations used in the value domain
	 * @param entityRefs
	 * @return List of unique associations found in Value Domain
	 */
	public List<String> getUniqueAssociationInVD(List<EntityReferenceWrapper> entityRefs){
		List<String> assns = new ArrayList<String>();
		for(EntityReferenceWrapper entityRef : entityRefs)
		{
			String assn = entityRef.getReferenceAssociation();
			if (!assns.contains(assn))
				assns.add(assn);
		}
		
		return assns;
	}
	
	/**
	 * Returns all unique coding schemes referenced in the value domain
	 * @param entityRefs
	 * @return List of Unique CodingSchemes in a Value Domain
	 */
	public List<String> getUniqueCodingSchemesInVD(List<EntityReferenceWrapper> entityRefs){
		List<String> csNames = new ArrayList<String>();
		for(EntityReferenceWrapper entityRef : entityRefs)
		{
			String csName = entityRef.getCodingSchemeReference();
			if (!csNames.contains(csName))
				csNames.add(csName);
		}		
		return csNames;
	}
	
	/**
	 * Populates and returns entity reference objects for each definitionEntry in value domain
	 * @param vdDef
	 * @return List of EntityReferenceWrapper
	 * @throws LBException
	 * @throws FindException
	 * @throws ServiceInitException
	 * @throws URISyntaxException
	 */
	public List<EntityReferenceWrapper> getEntityReferences(ValueDomainDefinition vdDef) 
			throws LBException, FindException, ServiceInitException, URISyntaxException {
		
		List<EntityReferenceWrapper> entityRefs = new ArrayList<EntityReferenceWrapper>();

		String defaultCS = vdDef.getDefaultCodingScheme();
		Iterator<DefinitionEntry> iter = vdDef.getDefinitionEntry().iterator();

		while (iter.hasNext()) {
			DefinitionEntry de = iter.next();
			EntityReferenceWrapper ec = new EntityReferenceWrapper();

			EntityReference er = de.getEntityReference();

			if (de.getCodingSchemeReference() != null) {
				ec.setCodingSchemeReference(de.getCodingSchemeReference().getCodingScheme());
			}

			if (ec.getCodingSchemeReference() == null)
				ec.setCodingSchemeReference(defaultCS);

			// try getting coding scheme information if there is value domain
			// reference
			if (ec.getCodingSchemeReference() == null && de.getValueDomainReference() != null) {
				String vdURI = de.getValueDomainReference().getValueDomainURI();
				if (StringUtils.isNotEmpty(vdURI))
					ec.setCodingSchemeReference(getCodingSchemeName(new URI(vdURI), er.getEntityCode()));
			}

			if (ec.getCodingSchemeReference() != null) {
				if (er != null) {
					ec.setEntityCode(er.getEntityCode());
					ec.setReferenceAssociation(er.getReferenceAssociation());
					ec.setLeafOnly(er.isLeafOnly());
					ec.setTransitiveClosure(er.isTransitiveClosure());
					ec.setTargetToSource(er.isTargetToSource());

					// Check if the associaion provided is associationName but
					// not the entityCode of an association
//					if (StringUtils.isNotEmpty(ec.getReferenceAssociation()))
//						ec.setReferenceAssociation(getAssociationEntityCode(ec.getCodingSchemeReference(), null, 
//								ec.getReferenceAssociation()));
					
					ec.setRuleOrder(de.getRuleOrder());
					ec.setDefinitionOperator(de.getOperator());
					
					if (!entityRefs.contains(ec))
						entityRefs.add(ec);
				}
			}
		}
		return entityRefs;
	}
	
	public VDServices getValueDomainServices() throws LBException {
		if (vds_ == null) {
			try {
				vds_ = getVDDefinitionService();
			} catch (ServiceInitException e) {
				md_.fatal("Failed while getting ValueDomainService ", e);
				throw new LBException(e.getMessage());
			}
		}
		return vds_;
	}

	public void setValueDomainServices(VDServices vds) {
		this.vds_ = vds;
	}

	public VDSServices getValueDomainsServices() throws ServiceInitException {
		if (vdss_ == null) {
			vdss_ = getVDSService();
		}
		return vdss_;
	}

	public void setValueDomainsServices(VDSServices vdss_) {
		this.vdss_ = vdss_;
	}

	public void setLexBIGService(LexBIGService lbs) {
		this.lbs_ = lbs;
	}

	public LexBIGService getLexBIGService() {
		if (lbs_ == null)
			lbs_ = LexBIGServiceImpl.defaultInstance();
		return lbs_;
	}
}