/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.loader.listener;

import java.util.Date;

import org.LexGrid.LexBIG.DataModel.InterfaceElements.types.ProcessState;
import org.LexGrid.persistence.dao.LexEvsDao;
import org.LexGrid.persistence.model.CodingScheme;
import org.LexGrid.persistence.properties.PropertiesFactory;
import org.lexgrid.loader.connection.LoaderConnectionManager;
import org.lexgrid.loader.connection.impl.LexEVSConnectionManager;
import org.lexgrid.loader.constants.LoaderConstants;
import org.lexgrid.loader.data.codingScheme.CodingSchemeNameSetter;
import org.lexgrid.loader.logging.LoggingBean;
import org.lexgrid.loader.setup.JobRepositoryManager;
import org.lexgrid.loader.staging.StagingManager;
import org.springframework.batch.core.ExitStatus;
import org.springframework.batch.core.JobExecution;
import org.springframework.batch.core.JobExecutionListener;

/**
 * The listener interface for receiving cleanup events.
 * The class that is interested in processing a cleanup
 * event implements this interface, and the object created
 * with that class is registered with a component using the
 * component's <code>addCleanupListener<code> method. When
 * the cleanup event occurs, that object's appropriate
 * method is invoked.
 * 
 * @see CleanupEvent
 */
public class CleanupListener extends LoggingBean implements JobExecutionListener {

	/** The connection manager. */
	private LoaderConnectionManager connectionManager = new LexEVSConnectionManager();
	
	/** The lex evs dao. */
	private LexEvsDao lexEvsDao;
	
	/** The coding scheme name setter. */
	private CodingSchemeNameSetter codingSchemeNameSetter;
	
	/** The staging manager. */
	private StagingManager stagingManager;
	
	/** The job repository manager. */
	private JobRepositoryManager jobRepositoryManager;
	
	private String prefix;
	
	private String database;
	
	/* (non-Javadoc)
	 * @see org.springframework.batch.core.JobExecutionListener#afterJob(org.springframework.batch.core.JobExecution)
	 */
	public void afterJob(JobExecution jobExecution) {
		
		if(jobExecution.getExitStatus().equals(ExitStatus.COMPLETED)){
			getLogger().info("Job completed, dropping tables");
			try {
				stagingManager.dropAllStagingDatabases();
				connectionManager.deactivate(getCurrentCodingSchemeUri(), getCurrentCodingSchemeVersion());
				connectionManager.initLoadedScheme(getCurrentCodingSchemeUri(), getCurrentCodingSchemeVersion());
			
				super.getLogger().getProcessStatus().setState(ProcessState.COMPLETED);
			} catch (Exception e) {
				throw new RuntimeException("Not all Staging Databases have been dropped.", e);
			}
		} else if(jobExecution.getExitStatus().getExitCode().equals(LoaderConstants.NON_RECOVERABLE)){
			getLogger().info("Non recoverable error in Job Processing. Job is not restartable, dropping all tables.");
			try {	
				String prefix = getParameterFromJobExecution(PropertiesFactory.PREFIX, jobExecution);
				String database = getParameterFromJobExecution(PropertiesFactory.DATABASE, jobExecution);
				
				connectionManager.cleanupFailedLoad(database, prefix);
				stagingManager.dropAllStagingDatabases();
				jobRepositoryManager.dropJobRepositoryDatabasesOnClose();
				
				super.getLogger().getProcessStatus().setState(ProcessState.FAILED);
			} catch (Exception e) {
				throw new RuntimeException("Not all Staging Databases have been cleaned up.", e);
			}
		}
		
		super.getLogger().getProcessStatus().setEndTime(new Date());
	}
	
	private String getParameterFromJobExecution(String parameter, JobExecution jobExecution){
		return jobExecution.getJobInstance().getJobParameters().getString(parameter);
	}

	/**
	 * Gets the staging manager.
	 * 
	 * @return the staging manager
	 */
	public StagingManager getStagingManager() {
		return stagingManager;
	}

	/**
	 * Sets the staging manager.
	 * 
	 * @param stagingManager the new staging manager
	 */
	public void setStagingManager(StagingManager stagingManager) {
		this.stagingManager = stagingManager;
	}
	
	/**
	 * Gets the current coding scheme uri.
	 * 
	 * @return the current coding scheme uri
	 * 
	 * @throws Exception the exception
	 */
	protected String getCurrentCodingSchemeUri() throws Exception {	
		return getCurrentCodingScheme().getCodingSchemeUri();
	}
	
	/**
	 * Gets the current coding scheme version.
	 * 
	 * @return the current coding scheme version
	 * 
	 * @throws Exception the exception
	 */
	protected String getCurrentCodingSchemeVersion() throws Exception {
		return getCurrentCodingScheme().getRepresentsVersion();
	}
	
	/**
	 * Gets the current coding scheme.
	 * 
	 * @return the current coding scheme
	 * 
	 * @throws Exception the exception
	 */
	protected CodingScheme getCurrentCodingScheme() throws Exception {
		return lexEvsDao.findById(CodingScheme.class, codingSchemeNameSetter.getCodingSchemeName());
	}

	/**
	 * Gets the connection manager.
	 * 
	 * @return the connection manager
	 */
	public LoaderConnectionManager getConnectionManager() {
		return connectionManager;
	}

	/**
	 * Sets the connection manager.
	 * 
	 * @param connectionManager the new connection manager
	 */
	public void setConnectionManager(LoaderConnectionManager connectionManager) {
		this.connectionManager = connectionManager;
	}

	/**
	 * Gets the lex evs dao.
	 * 
	 * @return the lex evs dao
	 */
	public LexEvsDao getLexEvsDao() {
		return lexEvsDao;
	}

	/**
	 * Sets the lex evs dao.
	 * 
	 * @param lexEvsDao the new lex evs dao
	 */
	public void setLexEvsDao(LexEvsDao lexEvsDao) {
		this.lexEvsDao = lexEvsDao;
	}
	
	/**
	 * Gets the coding scheme name setter.
	 * 
	 * @return the coding scheme name setter
	 */
	public CodingSchemeNameSetter getCodingSchemeNameSetter() {
		return codingSchemeNameSetter;
	}

	/**
	 * Sets the coding scheme name setter.
	 * 
	 * @param codingSchemeNameSetter the new coding scheme name setter
	 */
	public void setCodingSchemeNameSetter(
			CodingSchemeNameSetter codingSchemeNameSetter) {
		this.codingSchemeNameSetter = codingSchemeNameSetter;
	}

	/**
	 * Gets the job repository manager.
	 * 
	 * @return the job repository manager
	 */
	public JobRepositoryManager getJobRepositoryManager() {
		return jobRepositoryManager;
	}

	/**
	 * Sets the job repository manager.
	 * 
	 * @param jobRepositoryManager the new job repository manager
	 */
	public void setJobRepositoryManager(JobRepositoryManager jobRepositoryManager) {
		this.jobRepositoryManager = jobRepositoryManager;
	}

	public String getPrefix() {
		return prefix;
	}

	public void setPrefix(String prefix) {
		this.prefix = prefix;
	}

	public String getDatabase() {
		return database;
	}

	public void setDatabase(String database) {
		this.database = database;
	}

	public void beforeJob(JobExecution arg0) {
		//no-op -- handle this logging elsewhere	
	}	
}
